"""
TextWin, the window showing the text messages and info messages in poezio.
Can be locked, scrolled, has a separator, etc…
"""

import logging
from typing import Optional, List, Union

from poezio.windows.base_wins import Win
from poezio.text_buffer import TextBuffer

from poezio.config import config
from poezio.theming import to_curses_attr, get_theme
from poezio.ui.types import Message, BaseMessage
from poezio.ui.render import Line, ReactionLine, build_lines, write_pre

log = logging.getLogger(__name__)


class TextWin(Win):
    __slots__ = ('lines_nb_limit', 'pos', 'built_lines', 'lock', 'lock_buffer',
                 'separator_after', 'highlights', 'hl_pos',
                 'nb_of_highlights_after_separator')

    hl_pos: Optional[int]

    def __init__(self, lines_nb_limit: Optional[int] = None) -> None:
        Win.__init__(self)
        if lines_nb_limit is None:
            lines_nb_limit = config.getint('max_lines_in_memory')
        self.lines_nb_limit: int = lines_nb_limit
        self.pos = 0
        # Each new message is built and kept here.
        # on resize, we rebuild all the messages
        self.built_lines: List[Union[None, Line]] = []

        self.lock = False
        self.lock_buffer: List[Union[None, Line]] = []
        self.separator_after: Optional[BaseMessage] = None
        # the Lines of the highlights in that buffer
        self.highlights: List[Line] = []
        # the current HL position in that list NaN means that we’re not on
        # an hl. -1 is a valid position (it's before the first hl of the
        # list. i.e the separator, in the case where there’s no hl before
        # it.)
        self.hl_pos = None

        # Keep track of the number of hl after the separator.
        # This is useful to make “go to next highlight“ work after a “move to separator”.
        self.nb_of_highlights_after_separator = 0

    def toggle_lock(self) -> bool:
        if self.lock:
            self.release_lock()
        else:
            self.acquire_lock()
        return self.lock

    def acquire_lock(self) -> None:
        self.lock = True

    def release_lock(self) -> None:
        for line in self.lock_buffer:
            self.built_lines.append(line)
        self.lock = False

    def scroll_up(self, dist: int = 14) -> bool:
        pos = self.pos
        self.pos += dist
        if self.pos + self.height > len(self.built_lines):
            self.pos = len(self.built_lines) - self.height
            if self.pos < 0:
                self.pos = 0
        return self.pos != pos

    def scroll_down(self, dist: int = 14) -> bool:
        pos = self.pos
        self.pos -= dist
        if self.pos <= 0:
            self.pos = 0
        return self.pos != pos

    def build_new_message(self,
                          message: BaseMessage,
                          clean: bool = True,
                          timestamp: bool = False,
                          nick_size: int = 10,
                          reactions: bool = True) -> int:
        """
        Take one message, build it and add it to the list
        Return the number of lines that are built for the given
        message.
        """
        lines = build_lines(
            message, self.width, timestamp=timestamp, nick_size=nick_size
        )
        if self.lock:
            self.lock_buffer.extend(lines)
        else:
            self.built_lines.extend(lines)
        if not lines or not lines[0]:
            return 0
        if isinstance(message, Message) and message.highlight:
            self.highlights.append(lines[0])
            self.nb_of_highlights_after_separator += 1
            log.debug("Number of highlights after separator is now %s",
                      self.nb_of_highlights_after_separator)
        if clean:
            while len(self.built_lines) > self.lines_nb_limit:
                self.built_lines.pop(0)
        return len(lines)

    def refresh(self) -> None:
        log.debug('Refresh: %s', self.__class__.__name__)
        if self.height <= 0:
            return
        if self.pos == 0:
            lines = self.built_lines[-self.height:]
        else:
            lines = self.built_lines[-self.height - self.pos:-self.pos]
        with_timestamps = config.getbool("show_timestamps")
        nick_size = config.getint("max_nick_length")
        self._win.move(0, 0)
        self._win.erase()
        offset = 0
        for y, line in enumerate(lines):
            if isinstance(line, ReactionLine):
                msg = line.msg
                offset = msg.compute_offset(with_timestamps, nick_size)
                self.write_text(
                    y, offset,
                    line.txt,
                )
            elif isinstance(line, Line):
                msg = line.msg
                if line.start_pos == 0:
                    offset = write_pre(msg, self, with_timestamps, nick_size)
                elif y == 0:
                    offset = msg.compute_offset(with_timestamps,
                                                nick_size)
                self.write_text(
                    y, offset,
                    line.prepend + line.msg.txt[line.start_pos:line.end_pos])
            else:
                self.write_line_separator(y)
            if y != self.height - 1:
                self.addstr('\n')
        self._win.attrset(0)
        self._refresh()

    def write_text(self, y: int, x: int, txt: str) -> None:
        """
        write the text of a line.
        """
        self.addstr_colored(txt, y, x)

    def resize(self, height: int, width: int, y: int, x: int,
               room: Optional[TextBuffer] = None, force: bool = False) -> None:
        old_width: Optional[int]
        if hasattr(self, 'width'):
            old_width = self.width
        else:
            old_width = None
        self._resize(height, width, y, x)
        if room and (self.width != old_width or force):
            self.rebuild_everything(room)

        # reposition the scrolling after resize
        # (see #2450)
        buf_size = len(self.built_lines)
        if buf_size - self.pos < self.height:
            self.pos = buf_size - self.height
            if self.pos < 0:
                self.pos = 0

    def rebuild_everything(self, room: TextBuffer) -> None:
        self.built_lines = []
        with_timestamps = config.getbool('show_timestamps')
        nick_size = config.getint('max_nick_length')
        enable_reactions = config.getbool('enable_reactions')
        for message in room.filter_messages():
            self.build_new_message(
                message,
                clean=False,
                timestamp=with_timestamps,
                nick_size=nick_size,
                reactions=enable_reactions,
            )
            if self.separator_after is message:
                self.built_lines.append(None)
        while len(self.built_lines) > self.lines_nb_limit:
            self.built_lines.pop(0)

    def remove_line_separator(self) -> None:
        """
        Remove the line separator
        """
        log.debug('remove_line_separator')
        if None in self.built_lines:
            self.built_lines.remove(None)
            self.separator_after = None

    def add_line_separator(self, room: Optional[TextBuffer] = None) -> None:
        """
        add a line separator at the end of messages list
        room is a textbuffer that is needed to get the previous message
        (in case of resize)
        """
        if None not in self.built_lines:
            self.built_lines.append(None)
            self.nb_of_highlights_after_separator = 0
            log.debug("Resetting number of highlights after separator")
            if room and room.messages:
                self.separator_after = room.messages[-1]

    def write_line_separator(self, y: int) -> None:
        theme = get_theme()
        char = theme.CHAR_NEW_TEXT_SEPARATOR
        self.addnstr(y, 0, char * (self.width // len(char) - 1), self.width,
                     to_curses_attr(theme.COLOR_NEW_TEXT_SEPARATOR))

    def __del__(self) -> None:
        log.debug('** TextWin: deleting %s built lines',
                  (len(self.built_lines)))
        del self.built_lines

    def next_highlight(self) -> None:
        """
        Go to the next highlight in the buffer.
        (depending on which highlight was selected before)
        if the buffer is already positioned on the last, of if there are no
        highlights, scroll to the end of the buffer.
        """
        log.debug('Going to the next highlight…')
        if (not self.highlights or self.hl_pos is None
                or self.hl_pos >= len(self.highlights) - 1):
            self.hl_pos = None
            self.pos = 0
            return
        hl_size = len(self.highlights) - 1
        if self.hl_pos is not None and self.hl_pos < hl_size:
            self.hl_pos += 1
        else:
            self.hl_pos = hl_size
        log.debug("self.hl_pos = %s", self.hl_pos)
        hl = self.highlights[self.hl_pos]
        pos = None
        while not pos:
            try:
                pos = self.built_lines.index(hl)
            except ValueError:
                if isinstance(self.hl_pos, int):
                    del self.highlights[self.hl_pos]
                if not self.highlights:
                    self.hl_pos = None
                    self.pos = 0
                    return
                self.hl_pos = 0
                hl = self.highlights[0]
        self.pos = len(self.built_lines) - pos - self.height
        if self.pos < 0 or self.pos >= len(self.built_lines):
            self.pos = 0

    def previous_highlight(self) -> None:
        """
        Go to the previous highlight in the buffer.
        (depending on which highlight was selected before)
        if the buffer is already positioned on the first, or if there are no
        highlights, scroll to the end of the buffer.
        """
        log.debug('Going to the previous highlight…')
        if not self.highlights or self.hl_pos and self.hl_pos <= 0:
            self.hl_pos = None
            self.pos = 0
            return
        if self.hl_pos is None:
            self.hl_pos = len(self.highlights) - 1
        else:
            self.hl_pos -= 1
        log.debug("self.hl_pos = %s", self.hl_pos)
        hl = self.highlights[self.hl_pos]
        pos = None
        while not pos:
            try:
                pos = self.built_lines.index(hl)
            except ValueError:
                if self.hl_pos is not None:
                    del self.highlights[self.hl_pos]
                if not self.highlights:
                    self.hl_pos = None
                    self.pos = 0
                    return
                self.hl_pos = 0
                hl = self.highlights[0]
        self.pos = len(self.built_lines) - pos - self.height
        if self.pos < 0 or self.pos >= len(self.built_lines):
            self.pos = 0

    def scroll_to_separator(self) -> None:
        """
        Scroll to the first message after the separator.  If no
        separator is present, scroll to the first message of the window
        """
        if None in self.built_lines:
            self.pos = len(self.built_lines) - self.built_lines.index(
                None) - self.height + 1
            if self.pos < 0:
                self.pos = 0
        else:
            self.pos = len(self.built_lines) - self.height + 1
        # Chose a proper position (not too high)
        self.scroll_up(0)
        # Make “next highlight” work afterwards. This makes it easy to
        # review all the highlights since the separator was placed, in
        # the correct order.
        self.hl_pos = len(
            self.highlights) - self.nb_of_highlights_after_separator - 1
        log.debug("self.hl_pos = %s", self.hl_pos)

    def modify_message(self, old_id: str, message: Optional[BaseMessage]) -> None:
        """
        Find a message, and replace it with a new one
        (instead of rebuilding everything in order to correct a message)

        If "message" is not provided, the message is deleted and not replaced
        with a new one.
        """
        with_timestamps = config.getbool('show_timestamps')
        nick_size = config.getint('max_nick_length')
        enable_reactions = config.getbool('enable_reactions')
        for i in range(len(self.built_lines) - 1, -1, -1):
            current = self.built_lines[i]
            if current is not None and current.msg.identifier == old_id:
                index = i
                while (
                        index >= 0
                        and current is not None
                        and current.msg.identifier == old_id
                        ):
                    self.built_lines.pop(index)
                    index -= 1
                    if index >= 0:
                        current = self.built_lines[index]
                if message is not None:
                    index += 1
                    lines = build_lines(
                        message,
                        self.width,
                        timestamp=with_timestamps,
                        nick_size=nick_size,
                        reactions=enable_reactions,
                    )
                    for line in lines:
                        self.built_lines.insert(index, line)
                        index += 1
                break



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
Starting point of poezio. Launches both the Connection and Gui
"""

import sys
import os
import signal


def test_curses() -> bool:
    """
    Check if the system ncurses linked with python has unicode capabilities.
    """
    import curses
    if hasattr(curses, 'unget_wch'):
        return True
    print("""\
ERROR: The current python executable is linked with a ncurses version that \
has no unicode capabilities.

This could mean that:
    - python was built on a system where readline is linked against \
libncurses and not libncursesw
    - python was built without ncursesw headers available

Please file a bug for your distribution or fix that on your system and then \
recompile python.
Poezio is currently unable to read your input or draw its interface properly,\
 so it will now exit.""")
    return False


def test_env() -> bool:
    """
    Check if the environment has the right variables set
    """
    unset = set()
    for i in ('HOME', 'PATH', 'TERM'):
        if not os.environ.get(i):
            unset.add(i)
    if unset:
        variables = ', '.join('$%s' % i for i in unset)
        print('ERROR: %s not set' % variables)
        return False
    return True


def test_unicode() -> bool:
    from poezio import poopt
    try:
        poopt.wcswidth('✔')
    except UnicodeError:
        print("""\
ERROR: The current system is misconfigured for Unicode.

Check your locale setup, especially the $LANG environment variable and \
whether it matches a locale built on your system.  Also check that it is a \
.UTF-8 locale, and not using some legacy encoding.

Poezio is unable to display characters properly, so it will now exit.""")
        return False
    return True


def main() -> None:
    """
    Entry point.
    """

    if os.geteuid() == 0:
        sys.stdout.write("Please do not run poezio as root.\n")
        sys.exit(0)

    sys.stdout.write("\x1b]0;poezio\x07")
    sys.stdout.flush()
    from poezio.args import run_cmdline_args
    options, firstrun = run_cmdline_args()
    from poezio import config
    config.create_global_config(options.filename)
    config.setup_logging(options.debug)

    import logging
    logging.raiseExceptions = False

    if options.check_config:
        config.check_config()
        sys.exit(0)

    from poezio.asyncio_fix import monkey_patch_asyncio_slixmpp  # type: ignore
    monkey_patch_asyncio_slixmpp()

    from poezio import theming
    theming.update_themes_dir()

    from poezio.logger import logger
    logger.log_dir = config.LOG_DIR

    from poezio import roster
    roster.roster.reset()

    from poezio.core.core import Core

    signal.signal(signal.SIGINT, signal.SIG_IGN)  # ignore ctrl-c
    cocore = Core(options.custom_version, firstrun)
    signal.signal(signal.SIGUSR1, cocore.sigusr_handler)  # reload the config
    signal.signal(signal.SIGHUP, cocore.exit_from_signal)
    signal.signal(signal.SIGTERM, cocore.exit_from_signal)
    cocore.start()

    # Warning: asyncio must always be imported after the config. Otherwise
    # the asyncio logger will not follow our configuration and won't write
    # the tracebacks in the correct file, etc
    import asyncio
    loop = asyncio.get_event_loop()
    loop.set_exception_handler(cocore.loop_exception_handler)

    loop.add_reader(sys.stdin, cocore.on_input_readable)
    loop.add_signal_handler(signal.SIGWINCH, cocore.sigwinch_handler)
    cocore.xmpp.start()
    loop.run_forever()
    # We reach this point only when loop.stop() is called
    try:
        cocore.reset_curses()
    except Exception:
        logging.getLogger(__name__).debug('Unable to reset the screen')



"""
This plugin adds a :term:`/pacokick` command, which is a random kick.

Usage
-----

.. glossary::

    /pacokick

        Run the command in a room where you are a moderator to
        kick someone randomly.
"""

from random import choice
from poezio.tabs import MucTab

from poezio.plugin import BasePlugin


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_command(
            'pacokick',
            self.command_kick,
            usage='',
            help='Kick a random user.',
            short='Kick a random user')

    def command_kick(self, arg: str) -> None:
        tab = self.api.current_tab()
        if isinstance(tab, MucTab):
            kickable = list(
                filter(lambda x: x.affiliation in ('none', 'member'),
                       tab.users))
            if kickable:
                to_kick_user = choice(kickable)
                if to_kick_user:
                    to_kick = to_kick_user.nick
                    self.core.register_task(
                        tab.command_kick(to_kick + ' ' + arg)
                    )
            else:
                self.api.information('No one to kick :(', 'Info')



"""
Define the PluginConfig and Plugin classes, plus the SafetyMetaclass.
These are used in the plugin system added in poezio 0.7.5
(see plugin_manager.py)
"""

from typing import Any, Dict, Set, Optional, TYPE_CHECKING, Callable, Type
from configparser import RawConfigParser
from pathlib import Path
from poezio.timed_events import TimedEvent, DelayedEvent
from poezio.ui.types import Message
from poezio import tabs
from poezio.core.tabs import Tabs
from poezio.core.structs import Status, GenericCompletion
from poezio import config
import inspect
import traceback
import logging
log = logging.getLogger(__name__)


if TYPE_CHECKING:
    from poezio.core.core import Core
    from poezio.ui.types import BaseMessage
    from poezio.plugin_manager import PluginManager


class PluginConfig(config.Config):
    """
    Plugin configuration object.
    They are accessible inside the plugin with self.config
    and behave like the core Config object.
    """

    def __init__(self, filename: Path, module_name: str,
                 default: dict[Any, Any] | None = None) -> None:
        config.Config.__init__(self, filename, default=default)
        self.module_name = module_name
        self.default_section = module_name
        self.read()

    def get(self, option: str, default: config.ConfigValue | None = None,
            section: str | None = None):
        if not section:
            section = self.module_name
        return config.Config.get(self, option, default, section)

    def set(self, option: str, value: config.ConfigValue,
            section: str | None = None) -> None:
        if not section:
            section = self.module_name
        config.Config.set_and_save(self, option, value, section)

    def remove(self, option: str, section: str | None = None) -> tuple[str, str]:
        if not section:
            section = self.module_name
        return config.Config.remove_and_save(self, option, section)

    def read(self) -> None:
        """Read the config file"""
        RawConfigParser.read(self.configparser, str(self.file_name))
        if not self.has_section(self.module_name):
            self.add_section(self.module_name)

    def options(self, section: str | None = None) -> list[str]:
        """
            Return the options of the section
            If no section is given, it defaults to the plugin name.
        """
        if not section:
            section = self.module_name
        if not self.has_section(section):
            self.add_section(section)
        return config.Config.options(self, section)

    def write(self) -> bool:
        """Write the config to the disk"""
        try:
            with self.file_name.open('w') as fp:
                RawConfigParser.write(self.configparser, fp)
        except IOError:
            return False
        return True


class SafetyMetaclass(type):
    # A hack
    core: 'Core | None' = None

    @staticmethod
    def safe_func(f):
        def helper(*args, **kwargs):
            passthrough = kwargs.pop('passthrough', False)
            try:
                return f(*args, **kwargs)
            except Exception:
                if passthrough:
                    raise
                if inspect.stack()[1][1] == inspect.getfile(f):
                    raise
                elif SafetyMetaclass.core:
                    log.exception('Error in a plugin')
                    SafetyMetaclass.core.information(traceback.format_exc(),
                                                     'Error')
                    return None

        async def async_helper(*args, **kwargs):
            passthrough = kwargs.pop('passthrough', False)
            try:
                return await f(*args, **kwargs)
            except Exception:
                if passthrough:
                    raise
                if inspect.stack()[1][1] == inspect.getfile(f):
                    raise
                elif SafetyMetaclass.core:
                    log.exception('Error in a plugin')
                    SafetyMetaclass.core.information(traceback.format_exc(),
                                                     'Error')
                    return None
        if inspect.iscoroutinefunction(f):
            return async_helper
        return helper

    def __new__(meta, name, bases, class_dict):
        for k, v in class_dict.items():
            if inspect.isfunction(v):
                if k != '__init__' and k != 'init':
                    class_dict[k] = SafetyMetaclass.safe_func(v)
        return type.__new__(meta, name, bases, class_dict)


class PluginAPI:
    """
    The public API exposed to the plugins.
    Its goal is to limit the use of the raw Core object
    as much as possible.
    """

    def __init__(self, core: 'Core', plugin_manager: 'PluginManager', module_name: str) -> None:
        self.core = core
        self.plugin_manager = plugin_manager
        self.module_name = module_name

    def send_message(self, msg: str) -> bool:
        """
        Send a message to the current tab.

        :param str msg: The message to send.
        """
        return self.core.send_message(msg)

    def find_message(self, time: str, format: str) -> Optional[Message]:
        """Find a message from a timestamp.

        :param time: A timestamp used to find the message.
        :param format: The format of the timestamp.
        :returns: A Message or None
        """
        for message in reversed(self.core.get_conversation_messages()):
            if isinstance(message, Message) and \
               message.time.strftime(format) == time:
                return message
        return None

    def get_conversation_messages(self) -> 'list[BaseMessage]':
        """
        Get all the Messages of the current Tab.

        :returns: The list of :py:class:`text_buffer.Message` objects.
        :returns: None if the Tab does not inherit from ChatTab.
        :rtype: :py:class:`list`
        """
        return self.core.get_conversation_messages()

    def add_timed_event(self, event: DelayedEvent) -> None:
        """
        Schedule a timed event.

        :param timed_events.TimedEvent event: The timed event to schedule.
        """
        return self.core.add_timed_event(event)

    def remove_timed_event(self, event: DelayedEvent) -> None:
        """
        Unschedule a timed event.

        :param timed_events.TimedEvent event: The event to unschedule.
        """
        return self.core.remove_timed_event(event)

    def create_timed_event(self, *args, **kwargs) -> TimedEvent:
        """
        Create a timed event, but do not schedule it;
        :py:func:`~PluginAPI.add_timed_event` must be used for that.

        :param datetime.datetime date: The time at which the handler must be executed
        :param function callback: The handler that will be executed
        :param args: Optional arguments passed to the handler.
        :return: The created event.
        :rtype: :py:class:`timed_events.TimedEvent`
        """
        return TimedEvent(*args, **kwargs)

    def create_delayed_event(self, *args, **kwargs) -> DelayedEvent:
        """
        Create a delayed event, but do not schedule it;
        :py:func:`~PluginAPI.add_timed_event` must be used for that.

        A delayed event is a timed event with a delay from the time
        this function is called (instead of a datetime).

        :param int delay: The number of seconds to schedule the execution
        :param function callback: The handler that will be executed
        :param args: Optional arguments passed to the handler.
        :return: The created event.
        :rtype: :py:class:`timed_events.DelayedEvent`
        """
        return DelayedEvent(*args, **kwargs)

    def information(self, msg: str, typ: str = '') -> bool:
        """
        Display a new message in the information buffer.

        :param str msg: The message to display.
        :param str typ: The message type (e.g. Info, Error…)
        """
        return self.core.information(msg=msg, typ=typ)

    def current_tab(self) -> tabs.Tab:
        """
        Get the current Tab.

        :returns: The current tab.
        """
        return self.core.tabs.current_tab

    def get_status(self) -> Status:
        """
        Get the current user global status.

        :returns Status: The current status.
        """
        return self.core.get_status()

    def run_command(self, text: str) -> bool:
        """
        Run a command from the current tab.
        (a command starts with a /, if not, it’s a message)

        :param str line: The command to run.
        """
        return self.core.tabs.current_tab.execute_command(text)

    def all_tabs(self) -> Tabs:
        """
        Return a list of all opened tabs

        :returns list: The list of tabs.
        """
        return self.core.tabs

    def add_command(self,
                    name: str,
                    handler: Callable[[str], Any], help: str, *,
                    completion: GenericCompletion = None,
                    short: str = '', usage: str = '') -> None:
        """
        Add a global command.

        :param str name: The name of the command (/name)
        :param function handler: The function called when the command is run.
        :param str help: The complete help for that command.
        :param str short: A short description of the command.
        :param function completion: The completion function for that command
            (optional)
        :param str usage: A string showing the required and optional args
            of the command. Optional args should be surrounded by []
            and mandatory args should be surrounded by <>.

            Example string: "<server> [port]"

        :raises Exception: If the command already exists.
        """
        return self.plugin_manager.add_command(self.module_name, name, handler,
                                               help=help, completion=completion,
                                               short=short, usage=usage)

    def del_command(self, name: str) -> None:
        """
        Remove a global command.

        :param str name: The name of the command to remove.
            That command _must_ have been added by the same plugin
        """
        return self.plugin_manager.del_command(self.module_name, name)

    def add_key(self, key: str, handler: Callable[..., Any]) -> None:
        """
        Associate a global binding to a handler.

        :param str key: The curses representation of the binding.
        :param function handler: The function called when the binding is pressed.

        :raise Exception: If the binding is already present.
        """
        return self.plugin_manager.add_key(self.module_name, key, handler)

    def del_key(self, key: str) -> None:
        """
        Remove a global binding.

        :param str key: The binding to remove.
        """
        return self.plugin_manager.del_key(self.module_name, key)

    def add_tab_key(self, tab_type: Type[tabs.Tab], key: str,
                    handler: Callable[..., Any]) -> None:
        """
        Associate a binding to a handler, but only for a certain tab type.

        :param Tab tab_type: The type of tab to target.
        :param str key: The binding to add.
        :param function handler: The function called when the binding is pressed
        """
        return self.plugin_manager.add_tab_key(self.module_name, tab_type,
                                               key, handler)

    def del_tab_key(self, tab_type: Type[tabs.Tab], key: str) -> None:
        """
        Remove a binding added with add_tab_key

        :param tabs.Tab tab_type: The type of tab to target.
        :param str key: The binding to remove.
        """
        return self.plugin_manager.del_tab_key(self.module_name, tab_type, key)

    def add_tab_command(self,
                        tab_type: Type[tabs.Tab],
                        name: str,
                        handler: Callable[[str], Any], help: str, *,
                        completion: GenericCompletion = None,
                        short: str = '', usage: str = '') -> None:
        """
        Add a command to only one type of tab.

        :param tabs.Tab tab_type: The type of Tab to target.
        :param str name: The name of the command (/name)
        :param function handler: The function called when the command is run.
        :param str help: The complete help for that command.
        :param str short: A short description of the command.
        :param function completion: The completion function for that command
            (optional)
        :param str usage: A string showing the required and optional args
            of the command. Optional args should be surrounded by []
            and mandatory args should be surrounded by <>.

            Example string: "<server> [port]"

        :raise Exception: If the command already exists.
        """
        return self.plugin_manager.add_tab_command(
            self.module_name, tab_type, name, handler,
            help=help, completion=completion,
            short=short, usage=usage,
        )

    def del_tab_command(self, tab_type: Type[tabs.Tab], name: str) -> None:
        """
        Remove a tab-specific command.

        :param tabs.Tab tab_type: The type of tab to target.
        :param str name: The name of the command to remove.
            That command _must_ have been added by the same plugin
        """
        return self.plugin_manager.del_tab_command(self.module_name, tab_type, name)

    def add_event_handler(self, event_name: str, handler: Callable[..., Any],
                          priority: int = 50) -> None:
        """
        Add an event handler for a poezio event.

        :param str event_name: The event name.
        :param function handler: The handler function.
        :param int position: The position of that handler in the handler list.
            This is useful for plugins like OTR, which must be the last
            function called on the text.
            Defaults to 0.

        A complete list of those events can be found at
        https://doc.poez.io/dev/events.html
        """
        return self.plugin_manager.add_event_handler(
            self.module_name, event_name, handler, priority,
        )

    def del_event_handler(self, event_name: str, handler: Callable) -> None:
        """
        Remove a handler for a poezio event.

        :param str event_name: The name of the targeted event.
        :param function handler: The function to remove from the handlers.
        """
        return self.plugin_manager.del_event_handler(
            self.module_name, event_name, handler,
        )

    def add_slix_event_handler(self, event_name: str, handler: Callable[..., Any]) -> None:
        """
        Add an event handler for a slixmpp event.

        :param str event_name: The event name.
        :param function handler: The handler function.

        A list of the slixmpp events can be found here
        http://sleekxmpp.com/event_index.html
        """
        self.core.xmpp.add_event_handler(event_name, handler)

    def del_slix_event_handler(self, event_name: str, handler: Callable[..., Any]) -> None:
        """
        Remove a handler for a slixmpp event

        :param str event_name: The name of the targeted event.
        :param function handler: The function to remove from the handlers.
        """
        self.core.xmpp.del_event_handler(event_name, handler)


class BasePlugin(object, metaclass=SafetyMetaclass):
    """
    Class that all plugins derive from.
    """

    # Internal use only
    _unloading = False

    default_config: Optional[Dict[Any, Any]] = None
    dependencies: Set[str] = set()
    # This dict will get populated when the plugin is initialized
    refs: Dict[str, Any] = {}

    def __init__(self, name: str, plugin_api: 'PluginAPI', core: 'Core', plugins_conf_dir: Path) -> None:
        self.__name = name
        self.core = core
        # More hack; luckily we'll never have more than one core object
        SafetyMetaclass.core = core
        conf = plugins_conf_dir / (self.__name + '.cfg')
        try:
            self.config = PluginConfig(
                conf, self.__name, default=self.default_config)
        except Exception:
            log.debug('Error while creating the plugin config', exc_info=True)
            self.config = PluginConfig(conf, self.__name)
        self._api = plugin_api
        self.init()

    @property
    def name(self) -> str:
        """
        Get the name (module name) of the plugin.
        """
        return self.__name

    @property
    def api(self) -> 'PluginAPI':
        return self._api

    def init(self) -> None:
        """
        Method called at the creation of the plugin.

        Do not overwrite __init__ and use this instead.
        """
        pass

    def cleanup(self) -> None:
        """
        Called when the plugin is unloaded.

        Overwrite this if you want to erase or save things before the plugin is disabled.
        """
        pass

    def unload(self) -> None:
        self.cleanup()

    def add_command(self,
                    name: str,
                    handler: Callable[[str], Any], help: str, *,
                    completion: GenericCompletion = None,
                    short: str = '', usage: str = '') -> None:
        """
        Add a global command.

        :param str name: The name of the command (/name)
        :param function handler: The function called when the command is run.
        :param str help: The complete help for that command.
        :param str short: A short description of the command.
        :param function completion: The completion function for that command
            (optional)
        :param str usage: A string showing the required and optional args
            of the command. Optional args should be surrounded by []
            and mandatory args should be surrounded by <>.

            Example string: "<server> [port]"

        :raises Exception: If the command already exists.
        """
        return self.api.add_command(name, handler,
                                    help=help, completion=completion,
                                    short=short, usage=usage)

    def del_command(self, name: str) -> None:
        """
        Remove a global command.

        :param str name: The name of the command to remove.
            That command _must_ have been added by the same plugin
        """
        return self.api.del_command(name)

    def add_key(self, key: str, handler: Callable[..., Any]) -> None:
        """
        Associate a global binding to a handler.

        :param str key: The curses representation of the binding.
        :param function handler: The function called when the binding is pressed.

        :raise Exception: If the binding is already present.
        """
        return self.api.add_key(key, handler)

    def del_key(self, key: str) -> None:
        """
        Remove a global binding.

        :param str key: The binding to remove.
        """
        return self.api.del_key(key)

    def add_tab_key(self, tab_type: Type[tabs.Tab], key: str, handler: Callable[..., Any]) -> None:
        """
        Associate a binding to a handler, but only for a certain tab type.

        :param Tab tab_type: The type of tab to target.
        :param str key: The binding to add.
        :param function handler: The function called when the binding is pressed
        """
        return self.api.add_tab_key(tab_type, key, handler)

    def del_tab_key(self, tab_type: Type[tabs.Tab], key: str) -> None:
        """
        Remove a binding added with add_tab_key

        :param tabs.Tab tab_type: The type of tab to target.
        :param str key: The binding to remove.
        """
        return self.api.del_tab_key(tab_type, key)

    def add_tab_command(self,
                        tab_type: Type[tabs.Tab],
                        name: str,
                        handler: Callable[[str], Any], help: str, *,
                        completion: GenericCompletion = None,
                        short: str = '', usage: str = '') -> None:
        """
        Add a command to only one type of tab.

        :param tabs.Tab tab_type: The type of Tab to target.
        :param str name: The name of the command (/name)
        :param function handler: The function called when the command is run.
        :param str help: The complete help for that command.
        :param str short: A short description of the command.
        :param function completion: The completion function for that command
            (optional)
        :param str usage: A string showing the required and optional args
            of the command. Optional args should be surrounded by []
            and mandatory args should be surrounded by <>.

            Example string: "<server> [port]"

        :raise Exception: If the command already exists.
        """
        return self.api.add_tab_command(
            tab_type, name, handler,
            help=help, completion=completion,
            short=short, usage=usage,
        )

    def del_tab_command(self, tab_type: Type[tabs.Tab], name: str) -> None:
        """
        Remove a tab-specific command.

        :param tabs.Tab tab_type: The type of tab to target.
        :param str name: The name of the command to remove.
            That command _must_ have been added by the same plugin
        """
        return self.api.del_tab_command(tab_type, name)

    def add_event_handler(self, event_name: str, handler: Callable[..., Any],
                          priority: int = 50) -> None:
        """
        Add an event handler for a poezio event.

        :param str event_name: The event name.
        :param function handler: The handler function.
        :param int position: The position of that handler in the handler list.
            This is useful for plugins like OTR, which must be the last
            function called on the text.
            Defaults to 0.

        A complete list of those events can be found at
        https://doc.poez.io/dev/events.html
        """
        return self.api.add_event_handler(
            event_name, handler, priority,
        )

    def del_event_handler(self, event_name: str, handler: Callable[..., Any]) -> None:
        """
        Remove a handler for a poezio event.

        :param str event_name: The name of the targeted event.
        :param function handler: The function to remove from the handlers.
        """
        return self.api.del_event_handler(
            event_name, handler,
        )



"""
A tab listing the ad-hoc commands on a specific JID.  The user can
select one of them and start executing it, or just close the tab and do
nothing.
"""

import logging
from collections.abc import Iterator
from typing import Any, Callable, TYPE_CHECKING

from poezio.tabs import ListTab
from poezio.core.structs import Command

from slixmpp import JID, Iq
from slixmpp.plugins.xep_0030.stanza.items import DiscoItem

if TYPE_CHECKING:
    from poezio.core.core import Core

log = logging.getLogger(__name__)


class AdhocCommandsListTab(ListTab):
    plugin_commands: dict[str, Command] = {}
    plugin_keys: dict[str, Callable[..., Any]] = {}

    def __init__(self, core: 'Core', jid: JID) -> None:
        ListTab.__init__(
            self, core, jid.full, "“Enter”: execute selected command.",
            'Ad-hoc commands of JID %s (Loading)' % jid, [('Node', 0),
                                                          ('Description', 1)])
        self.key_func['^M'] = self.execute_selected_command

    def execute_selected_command(self) -> None:
        if not self.listview:
            return
        selected_row = self.listview.get_selected_row()
        if not selected_row:
            return
        node, name, jid = selected_row
        session = {
            'next': self.core.handler.next_adhoc_step,
            'error': self.core.handler.adhoc_error
        }
        self.core.xmpp.plugin['xep_0050'].start_command(jid, node, session)

    def get_columns_sizes(self) -> dict[str, int]:
        return {
            'Node': int(self.width * 3 / 8),
            'Description': int(self.width * 5 / 8)
        }

    def on_list_received(self, iq: Iq) -> None:
        """
        Fill the listview with the value from the received iq
        """
        if iq['type'] == 'error':
            self.set_error(iq['error']['type'], iq['error']['code'],
                           iq['error']['text'])
            return

        def get_items() -> Iterator[DiscoItem]:
            substanza = iq['disco_items']
            for item in substanza['substanzas']:
                if isinstance(item, DiscoItem):
                    yield item

        items = [(item['node'], item['name'] or '', item['jid'])
                 for item in get_items()]
        self.listview.set_lines(items)
        self.info_header.message = 'Ad-hoc commands of JID %s' % self.name
        if self.core.tabs.current_tab is self:
            self.refresh()
        else:
            self.state = 'highlight'
            self.refresh_tab_win()
        self.core.doupdate()



"""
Tests for the TextBuffer class
"""
from pytest import fixture, raises

from poezio.text_buffer import (
    CorrectionError,
    DeletionError,
    TextBuffer,
)
from poezio.user import User

from poezio.ui.types import (
    Message,
    MucJoin,
    MucLeave,
    BaseMessage,
    MucOwnJoinMessage,
    MucOwnLeaveMessage,
)
from poezio.xhtml import clean_text

from slixmpp import JID


@fixture(scope='function')
def buf2048():
    return TextBuffer(2048)


@fixture(scope='function')
def msgs_nojoin():
    msg1 = Message('1', 'q')
    msg2 = Message('2', 's')
    leave = MucOwnLeaveMessage('leave')
    return [msg1, msg2, leave]


@fixture(scope='function')
def msgs_noleave():
    join = MucOwnJoinMessage('join')
    msg3 = Message('3', 'd')
    msg4 = Message('4', 'f')
    return [join, msg3, msg4]


@fixture(scope='function')
def msgs_doublejoin():
    join = MucOwnJoinMessage('join')
    msg1 = Message('1', 'd')
    msg2 = Message('2', 'f')
    join2 = MucOwnJoinMessage('join')
    return [join, msg1, msg2, join2]


@fixture(scope='function')
def msgs_join_leave_nomsg():
    msg1 = Message('1', 'd')
    msg2 = Message('2', 'f')
    join = MucOwnJoinMessage('join')
    leave = MucOwnLeaveMessage('leave')
    join2 = MucOwnJoinMessage('join')
    return [msg1, msg2, join, leave, join2]


def test_last_message(buf2048):
    msg = BaseMessage('toto')
    buf2048.add_message(BaseMessage('titi'))
    buf2048.add_message(msg)
    assert buf2048.last_message is msg


def test_message_nb_limit():
    buf = TextBuffer(5)
    for i in range(10):
        buf.add_message(BaseMessage("%s" % i))
    assert len(buf.messages) == 5


def test_find_gap(buf2048, msgs_noleave):
    msg1 = Message('1', 'q')
    msg2 = Message('2', 's')
    leave = MucOwnLeaveMessage('leave')
    join = MucOwnJoinMessage('join')
    msg3 = Message('3', 'd')
    msg4 = Message('4', 'f')
    msgs = [msg1, msg2, leave, join, msg3, msg4]
    for msg in msgs:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert gap.leave_message == leave
    assert gap.join_message == join
    assert gap.last_timestamp_before_leave == msg2.time
    assert gap.first_timestamp_after_join == msg3.time


def test_find_gap_leavejoin(buf2048, msgs_join_leave_nomsg):
    for msg in msgs_join_leave_nomsg:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert gap.leave_message == msgs_join_leave_nomsg[3]
    assert gap.join_message == msgs_join_leave_nomsg[4]
    assert gap.last_timestamp_before_leave == msgs_join_leave_nomsg[3].time


def test_find_gap_doublejoin(buf2048, msgs_doublejoin):
    for msg in msgs_doublejoin:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert gap.leave_message == msgs_doublejoin[2]
    assert gap.join_message == msgs_doublejoin[3]


def test_find_gap_doublejoin_no_msg(buf2048):
    join1 = MucOwnJoinMessage('join')
    join2 = MucOwnJoinMessage('join')
    for msg in [join1, join2]:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert gap.leave_message is join1
    assert gap.join_message is join2


def test_find_gap_already_filled(buf2048):
    msg1 = Message('1', 'q')
    msg2 = Message('2', 's')
    leave = MucOwnLeaveMessage('leave')
    msg5 = Message('5', 'g')
    msg6 = Message('6', 'h')
    join = MucOwnJoinMessage('join')
    msg3 = Message('3', 'd')
    msg4 = Message('4', 'f')
    msgs = [msg1, msg2, leave, msg5, msg6, join, msg3, msg4]
    for msg in msgs:
        buf2048.add_message(msg)
    assert buf2048.find_last_gap_muc() is None


def test_find_gap_noleave(buf2048, msgs_noleave):
    for msg in msgs_noleave:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert gap.leave_message is None
    assert gap.last_timestamp_before_leave is None
    assert gap.join_message == msgs_noleave[0]
    assert gap.first_timestamp_after_join == msgs_noleave[1].time


def test_find_gap_nojoin(buf2048, msgs_nojoin):
    for msg in msgs_nojoin:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert gap.leave_message == msgs_nojoin[-1]
    assert gap.join_message is None
    assert gap.last_timestamp_before_leave == msgs_nojoin[1].time


def test_get_gap_index(buf2048):
    msg1 = Message('1', 'q')
    msg2 = Message('2', 's')
    leave = MucOwnLeaveMessage('leave')
    join = MucOwnJoinMessage('join')
    msg3 = Message('3', 'd')
    msg4 = Message('4', 'f')
    msgs = [msg1, msg2, leave, join, msg3, msg4]
    for msg in msgs:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert buf2048.get_gap_index(gap) == 3


def test_get_gap_index_doublejoin(buf2048, msgs_doublejoin):
    for msg in msgs_doublejoin:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert buf2048.get_gap_index(gap) == 3


def test_get_gap_index_doublejoin_no_msg(buf2048):
    join1 = MucOwnJoinMessage('join')
    join2 = MucOwnJoinMessage('join')
    for msg in [join1, join2]:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert buf2048.get_gap_index(gap) == 1


def test_get_gap_index_nojoin(buf2048, msgs_nojoin):
    for msg in msgs_nojoin:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert buf2048.get_gap_index(gap) == 3


def test_get_gap_index_noleave(buf2048, msgs_noleave):
    for msg in msgs_noleave:
        buf2048.add_message(msg)
    gap = buf2048.find_last_gap_muc()
    assert buf2048.get_gap_index(gap) == 0


def test_add_history_messages(buf2048):
    msg1 = Message('1', 'q')
    msg2 = Message('2', 's')
    leave = MucOwnLeaveMessage('leave')
    join = MucOwnJoinMessage('join')
    msg3 = Message('3', 'd')
    msg4 = Message('4', 'f')
    msgs = [msg1, msg2, leave, join, msg3, msg4]
    for msg in msgs:
        buf2048.add_message(msg)
    msg5 = Message('5', 'g')
    msg6 = Message('6', 'h')
    gap = buf2048.find_last_gap_muc()
    buf2048.add_history_messages([msg5, msg6], gap=gap)
    assert buf2048.messages == [msg1, msg2, leave, msg5, msg6, join, msg3, msg4]


def test_add_history_empty(buf2048):
    msg1 = Message('1', 'q')
    msg2 = Message('2', 's')
    msg3 = Message('3', 'd')
    msg4 = Message('4', 'f')
    buf2048.add_message(msg1)
    buf2048.add_history_messages([msg2, msg3, msg4])
    assert buf2048.messages == [msg2, msg3, msg4, msg1]


def test_modify_message(buf2048):
    user = User("toto", "owner", "", "xa", "moderator", JID("toto@titi"), color="red")
    msg1 = Message('1', 'q')
    msg2 = Message('2', 's', identifier="2", jid=JID("a@b"))
    msg3 = Message('3', 'd', identifier="3", user=user)
    msg4 = Message('4', 'f')
    msgs = [msg1, msg2, msg3, msg4]
    for msg in msgs:
        buf2048.add_message(msg)

    buf2048.modify_message("coucou", "2", "repl", jid=JID("a@b"))
    assert buf2048.messages[0].txt == "1"
    assert buf2048.messages[1].txt == "coucou"
    assert buf2048.messages[2].txt == "3"
    assert buf2048.messages[3].txt == "4"

    buf2048.modify_message("coucou 2", "3", "repl2", user=user)
    assert buf2048.messages[0].txt == "1"
    assert buf2048.messages[1].txt == "coucou"
    assert buf2048.messages[2].txt == "coucou 2"
    assert buf2048.messages[3].txt == "4"

    buf2048.modify_message("coucou 3", "repl", "repl3", jid=JID("a@b"))
    assert buf2048.messages[0].txt == "1"
    assert buf2048.messages[1].txt == "coucou 3"
    assert buf2048.messages[2].txt == "coucou 2"
    assert buf2048.messages[3].txt == "4"

    with raises(CorrectionError):
        buf2048.modify_message("coucou 4", "missing", "repl4", jid=JID("a@b"))

    with raises(CorrectionError):
        buf2048.modify_message("coucou 4", "repl3", "repl4", user=user)

    with raises(CorrectionError):
        buf2048.modify_message("coucou 4", "repl3", "repl4")


def test_delete_message(buf2048):
    user = User("toto", "owner", "", "xa", "moderator", JID("toto@titi"), color="red")
    user2 = User("titi", "user", "", "xa", "moderator", JID("titi@toto"), color="blue")
    msg1 = Message('1', 'q', identifier="1")
    msg2 = Message('2', 's', identifier="2", jid=JID("a@b"))
    msg3 = Message('3', 'd', identifier="3", user=user)
    msg4 = Message('4', 'f', identifier="4")
    msg5 = Message('5', 'g', identifier="5")
    msg6 = Message('6', 'h', identifier="6")
    msgs = [msg1, msg2, msg3, msg4, msg5, msg6]
    for msg in msgs:
        buf2048.add_message(msg)

    buf2048.delete_message("4", moderation=True)
    assert buf2048.messages[3].txt == "5"

    buf2048.delete_message("5", moderation=True, tombstone="deleted by toto")
    assert "deleted by toto" in buf2048.messages[3].txt

    with raises(DeletionError):
        buf2048.delete_message("2", jid=JID("c@d"))

    with raises(DeletionError):
        buf2048.delete_message("3", user=user2)

    with raises(DeletionError):
        buf2048.delete_message("3")

    assert not buf2048.delete_message("none")

    buf2048.delete_message("2", jid=JID("a@b"))
    assert buf2048.messages[1].txt == "3"

    buf2048.delete_message("3", user=user)
    assert "deleted by" in buf2048.messages[1].txt


def test_reactions(buf2048):
    buf2048.add_message(Message('body', 'nick', identifier="0"))
    buf2048.add_message(Message('body2', 'nick2', identifier="1"))
    buf2048.add_message(Message('body3', 'nick3', identifier="2"))
    assert buf2048.react(to='1', by='toto', reactions={'a', 'b'})
    assert buf2048.react(to='1', by='titi', reactions={'a', 'b', 'c'})
    assert buf2048.react(to='1', by='tata', reactions={'a', 'b', 'c', 'd'})

    assert buf2048.messages[1].reactions['a'] == {'toto', 'titi', 'tata'}
    assert buf2048.messages[1].reactions['b'] == {'toto', 'titi', 'tata'}
    assert buf2048.messages[1].reactions['c'] == {'titi', 'tata'}

    # Remove a reaction now
    assert buf2048.react(to='1', by='tata', reactions={'a', 'b'})

    assert buf2048.messages[1].reactions['a'] == {'toto', 'titi', 'tata'}
    assert buf2048.messages[1].reactions['b'] == {'toto', 'titi', 'tata'}
    assert buf2048.messages[1].reactions['c'] == {'titi'}

    # Message not found
    assert not buf2048.react(to='5', by='tata', reactions={'p'})
    # No change
    assert not buf2048.react(to='1', by='tata', reactions={'a', 'b'})


def test_filters(buf2048):
    buf2048.add_message(MucLeave("tutu"))
    buf2048.add_message(Message('body', 'nick', identifier="0"))
    buf2048.add_message(MucJoin("toto"))
    buf2048.add_message(Message('body2', 'nick2', identifier="1"))
    buf2048.add_message(MucLeave("tata"))
    buf2048.add_message(Message('body3', 'nick3', identifier="2"))
    buf2048.add_message(MucJoin("titi"))

    buf2048.filter_classes((MucLeave, MucJoin))
    filtered = buf2048.filter_messages()
    assert len(filtered) == 3
    assert filtered[0].txt == 'body'
    assert filtered[1].txt == 'body2'
    assert filtered[2].txt == 'body3'

    buf2048.filter_classes((MucLeave,))
    filtered = buf2048.filter_messages()
    assert len(filtered) == 5
    assert filtered[0].txt == 'body'
    assert clean_text(filtered[1].txt) == 'toto'
    assert filtered[2].txt == 'body2'
    assert filtered[3].txt == 'body3'
    assert clean_text(filtered[4].txt) == 'titi'

    buf2048.filter_classes((MucJoin,))
    filtered = buf2048.filter_messages()
    assert len(filtered) == 5
    assert clean_text(filtered[0].txt) == 'tutu'
    assert filtered[1].txt == 'body'
    assert filtered[2].txt == 'body2'
    assert clean_text(filtered[3].txt) == 'tata'
    assert filtered[4].txt == 'body3'



"""
Marquee plugin: replicate the html <marquee/> tag with message corrections.

Usage of this plugin is not recommended.

Commands
--------

.. glossary::

    /marquee <text>
        Send the following text with <marquee/> behavior

Configuration
-------------

.. glossary::
    :sorted:

    refresh [marquee plugin]
        **Default:** ``1``

        Interval between each correction (the closest to 0 is the fastest)

    total_duration
        **Default:** ``30``

        Total duration of the animation.

    padding
        **Default:** ``20``

        Padding to use to move the text.


"""
from poezio.plugin import BasePlugin
from poezio import tabs
from poezio import xhtml
from poezio.decorators import command_args_parser
from slixmpp import JID


def move(text: str, step: int, spacing: int) -> str:
    new_text = text + ("\u00A0" * spacing)
    return new_text[-(step % len(new_text)):] + new_text[:-(
        step % len(new_text))]


class Plugin(BasePlugin):
    default_config = {
        "marquee": {
            "refresh": 1.0,
            "total_duration": 30,
            "padding": 20
        }
    }

    def init(self) -> None:
        for tab_t in [tabs.MucTab, tabs.DynamicConversationTab, tabs.StaticConversationTab, tabs.PrivateTab]:
            self.add_tab_command(
                tab_t, 'marquee', self.command_marquee,
                'Replicate the <marquee/> behavior in a message')

    @command_args_parser.raw
    async def command_marquee(self, args: str) -> None:
        if not args:
            return
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.ChatTab):
            return
        args = xhtml.clean_text(xhtml.convert_simple_to_full_colors(args))
        await tab.command_say(args)
        is_muctab = isinstance(tab, tabs.MucTab)
        if tab.last_sent_message is None:
            return
        msg_id = tab.last_sent_message["id"]
        jid = tab.jid

        event = self.api.create_delayed_event(
            self.config.get("refresh"), self.delayed_event, jid, args, msg_id,
            1, 0, is_muctab)
        self.api.add_timed_event(event)

    def delayed_event(self, jid: JID, body: str, msg_id: str,
                      step: int, duration: int, is_muctab: bool) -> None:
        if duration >= self.config.get("total_duration"):
            return
        message = self.core.xmpp.make_message(jid)
        message["type"] = "groupchat" if is_muctab else "chat"
        message["body"] = move(body, step, self.config.get("padding"))
        message["replace"]["id"] = msg_id
        message.send()
        event = self.api.create_delayed_event(
            self.config.get("refresh"), self.delayed_event, jid, body,
            msg_id, step + 1, duration + self.config.get("refresh"),
            is_muctab)
        self.api.add_timed_event(event)



"""
Module defining the Core class, which is the central orchestrator
of poezio and contains the main loop, the list of tabs, sets the state
of everything; it also contains global commands, completions and event
handlers but those are defined in submodules in order to avoir cluttering
this file.
"""
from __future__ import annotations

import logging
import asyncio
import curses
import os
import shlex
import sys
import shutil
import time
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Concatenate,
    Coroutine,
    Dict,
    List,
    Optional,
    ParamSpec,
    Set,
    Unpack,
    Tuple,
    Type,
    TypeVar,
)
from xml.etree import ElementTree as ET
from types import TracebackType
from asyncio.events import AbstractEventLoop
from pathlib import Path

from slixmpp import Iq, JID, InvalidJID, Message, Presence
from slixmpp.util import FileSystemPerJidCache
from slixmpp.xmlstream.xmlstream import InvalidCABundle
from slixmpp.xmlstream.handler import Callback
from slixmpp.exceptions import IqError, IqTimeout
from slixmpp.stanza import StreamError
from slixmpp.plugins.xep_0004 import Form

from poezio import connection
from poezio import decorators
from poezio import events
from poezio import theming
from poezio import timed_events
from poezio import windows
from poezio import utils

from poezio.bookmarks import (
    BookmarkList,
    Bookmark,
)
from poezio.tabs import (
    Tab, XMLTab, ChatTab, ConversationTab, PrivateTab, MucTab, OneToOneTab,
    GapTab, RosterInfoTab, StaticConversationTab, DataFormsTab,
    DynamicConversationTab, STATE_PRIORITY
)
from poezio.common import get_error_message
from poezio.config import config, USE_DEFAULT_SECTION
from poezio.contact import Contact, Resource
from poezio.daemon import Executor
from poezio.fifo import Fifo
from poezio.logger import logger
from poezio.plugin_manager import PluginManager
from poezio.roster import roster, RosterGroup
from poezio.size_manager import SizeManager
from poezio.user import User
from poezio.text_buffer import TextBuffer
from poezio.timed_events import DelayedEvent
from poezio import keyboard, xdg

from poezio.core.completions import CompletionCore
from poezio.core.tabs import Tabs
from poezio.core.commands import CommandCore
from poezio.core.command_defs import get_commands
from poezio.core.handlers import HandlerCore
from poezio.core.structs import (
    Command,
    CommandKwargs,
    Status,
    POSSIBLE_SHOW,
)

from poezio.ui.types import (
    BaseMessage,
    PersistentInfoMessage,
    UIMessage,
)


log = logging.getLogger(__name__)

P = ParamSpec('P')
S = TypeVar('S')
T = TypeVar('T', bound=Tab)
U = TypeVar('U')


class Core:
    """
    “Main” class of poezion
    """

    custom_version: str
    firstrun: bool
    completion: CompletionCore
    command: CommandCore
    handler: HandlerCore
    bookmarks: BookmarkList
    status: Status
    commands: Dict[str, Command]
    room_number_jump: List[str]
    initial_joins: List[JID]
    pending_invites: Dict[str, str]
    configuration_change_handlers: dict[str, list[Callable[..., None]]]
    tab_configuration_change_handlers: dict[str, list[Callable[..., None]]]
    own_nick: str
    connection_time: float
    xmpp: connection.Connection
    avatar_cache: FileSystemPerJidCache
    plugins_autoloaded: bool
    previous_tab_nb: int
    tabs: Tabs
    size: SizeManager
    plugin_manager: PluginManager
    events: events.EventHandler
    legitimate_disconnect: bool
    information_buffer: TextBuffer
    information_win_size: int
    stdscr: Optional[curses.window]
    xml_buffer: TextBuffer
    xml_tab: Optional[XMLTab]
    last_stream_error: Optional[Tuple[float, StreamError]]
    remote_fifo: Optional[Fifo]
    key_func: KeyDict
    tab_win: windows.GlobalInfoBar
    left_tab_win: Optional[windows.VerticalGlobalInfoBar]
    _tasks: list[asyncio.Task[Any]]

    def __init__(self, custom_version: str, firstrun: bool):
        self.completion = CompletionCore(self)
        self.command = CommandCore(self)
        self.handler = HandlerCore(self)
        self.firstrun = firstrun
        self._tasks = []
        # All uncaught exception are given to this callback, instead
        # of being displayed on the screen and exiting the program.
        sys.excepthook = self.on_exception
        self.connection_time = time.time()
        self.last_stream_error = None
        self.stdscr = None
        status = config.getstr('status')
        status = POSSIBLE_SHOW.get(status) or ''
        self.status = Status(show=status, message=config.getstr('status_message'))
        self.custom_version = custom_version
        self.xmpp = connection.Connection(custom_version)
        self.xmpp.core = self
        self.keyboard = keyboard.Keyboard()
        roster.set_node(self.xmpp.client_roster)
        decorators.refresh_wrapper.core = self
        self.bookmarks = BookmarkList()
        self.remote_fifo = None
        self.avatar_cache = FileSystemPerJidCache(
            str(xdg.CACHE_HOME), 'avatars', binary=True)
        # a unique buffer used to store global information
        # that are displayed in almost all tabs, in an
        # information window.
        self.information_buffer = TextBuffer()
        self.information_win_size = config.getint('info_win_height', section='var')

        # Whether the XML tab is opened
        self.xml_tab = None
        self.xml_buffer = TextBuffer()

        self.plugins_autoloaded = False
        self.plugin_manager = PluginManager(self)
        self.events = events.EventHandler()
        self.events.add_event_handler('tab_change', self.on_tab_change)

        self.tabs = Tabs(self.events, GapTab())
        self.previous_tab_nb = 0

        self.own_nick: str = (
            config.getstr('default_nick') or self.xmpp.boundjid.user or
            os.environ.get('USER') or 'poezio_user'
        )

        self.size = SizeManager(self)

        # Set to True whenever we consider that we have been disconnected
        # from the server because of a legitimate reason (bad credentials,
        # or explicit disconnect from the user for example), in that case we
        # should not try to auto-reconnect, even if auto_reconnect is true
        # in the user config.
        self.legitimate_disconnect = False

        # global commands, available from all tabs
        # a command is tuple of the form:
        # (the function executing the command. Takes a string as argument,
        #  a string representing the help message,
        #  a completion function, taking a Input as argument. Can be None)
        #  The completion function should return True if a completion was
        #  made ; False otherwise
        self.commands = {}
        self.register_initial_commands()

        # We are invisible
        if not config.get('send_initial_presence'):
            del self.commands['status']
            del self.commands['show']

        # A list of integers. For example if the user presses Alt+j, 2, 1,
        # we will insert 2, then 1 in that list, and we will finally build
        # the number 21 and use it with command_win, before clearing the
        # list.
        self.room_number_jump = []
        self.key_func = KeyDict()
        # Key bindings associated with handlers
        # and pseudo-keys used to map actions below.
        key_func: dict[str, Callable[..., Any]] = {
            "KEY_PPAGE": self.scroll_page_up,
            "KEY_NPAGE": self.scroll_page_down,
            "^B": self.scroll_line_up,
            "^F": self.scroll_line_down,
            "^X": self.scroll_half_down,
            "^S": self.scroll_half_up,
            "KEY_F(5)": self.rotate_rooms_left,
            "^P": self.rotate_rooms_left,
            "M-[-D": self.rotate_rooms_left,
            "M-[1;3D": self.rotate_rooms_left,
            'kLFT3': self.rotate_rooms_left,
            "KEY_F(6)": self.rotate_rooms_right,
            "^N": self.rotate_rooms_right,
            "M-[-C": self.rotate_rooms_right,
            "M-[1;3C": self.rotate_rooms_right,
            'kRIT3': self.rotate_rooms_right,
            "KEY_F(4)": self.toggle_left_pane,
            "KEY_F(7)": self.shrink_information_win,
            "KEY_F(8)": self.grow_information_win,
            "KEY_RESIZE": self.call_for_resize,
            'M-e': self.go_to_important_room,
            'M-r': self.go_to_roster,
            'M-z': self.go_to_previous_tab,
            '^L': self.full_screen_redraw,
            'M-j': self.go_to_room_number,
            'M-D': self.scroll_info_up,
            'M-C': self.scroll_info_down,
            'M-k': self.escape_next_key,
            ######## actions mappings ##########
            '_noop': lambda *args, **kwargs: None,
            '_bookmark': self.command.bookmark,
            '_bookmark_local': self.command.bookmark_local,
            '_close_tab': self.close_tab,
            '_disconnect': self.disconnect,
            '_quit': self.command.quit,
            '_redraw_screen': self.full_screen_redraw,
            '_reload_theme': self.command.theme,
            '_remove_bookmark': self.command.remove_bookmark,
            '_room_left': self.rotate_rooms_left,
            '_room_right': self.rotate_rooms_right,
            '_show_roster': self.go_to_roster,
            '_scroll_down': self.scroll_page_down,
            '_scroll_up': self.scroll_page_up,
            '_scroll_info_up': self.scroll_info_up,
            '_scroll_info_down': self.scroll_info_down,
            '_server_cycle': self.command.server_cycle,
            '_show_bookmarks': self.command.bookmarks,
            '_show_important_room': self.go_to_important_room,
            '_show_invitations': self.command.invitations,
            '_show_plugins': self.command.plugins,
            '_show_xmltab': self.command.xml_tab,
            '_toggle_pane': self.toggle_left_pane,
            '_toggle_presences_local': lambda: self.toggle_show_presences(local=True),
            '_toggle_presences_global': lambda: self.toggle_show_presences(local=False),
            "_go_to_room_name": self.go_to_room_name,
            ###### status actions ######
            '_available': lambda: self.command.status('available'),
            '_away': lambda: self.command.status('away'),
            '_chat': lambda: self.command.status('chat'),
            '_dnd': lambda: self.command.status('dnd'),
            '_xa': lambda: self.command.status('xa'),
            ##### Custom actions ########
        }
        self.key_func.update(key_func)
        self.key_func.try_execute = self.try_execute

        # Add handlers
        xmpp_event_handlers: List[Tuple[str, Callable[..., Any]]] = [
            ('attention', self.handler.on_attention),
            ('carbon_received', self.handler.on_carbon_received),
            ('carbon_sent', self.handler.on_carbon_sent),
            ('changed_status', self.handler.on_presence),
            ('chatstate_active', self.handler.on_chatstate_active),
            ('chatstate_composing', self.handler.on_chatstate_composing),
            ('chatstate_gone', self.handler.on_chatstate_gone),
            ('chatstate_inactive', self.handler.on_chatstate_inactive),
            ('chatstate_paused', self.handler.on_chatstate_paused),
            ('connected', self.handler.on_connected),
            ('connection_failed', self.handler.on_failed_connection),
            ('disconnected', self.handler.on_disconnected),
            ('reconnect_delay', self.handler.on_reconnect_delay),
            ('failed_all_auth', self.handler.on_failed_all_auth),
            ('got_offline', self.handler.on_got_offline),
            ('got_online', self.handler.on_got_online),
            ('groupchat_config_status', self.handler.on_status_codes),
            ('groupchat_decline', self.handler.on_groupchat_decline),
            ('groupchat_direct_invite',
             self.handler.on_groupchat_direct_invitation),
            ('groupchat_invite', self.handler.on_groupchat_invitation),
            ('groupchat_message', self.handler.on_groupchat_message),
            ('groupchat_presence', self.handler.on_groupchat_presence),
            ('groupchat_subject', self.handler.on_groupchat_subject),
            ('http_confirm', self.handler.http_confirm),
            ('message', self.handler.on_message),
            ('message_encryption', self.handler.on_encrypted_message),
            ('message_error', self.handler.on_error_message),
            ('message_xform', self.handler.on_data_form),
            ('message_retract', self.handler.on_message_retract),
            ('muc_ping_changed', self.handler.muc_ping_result),
            ('no_auth', self.handler.on_no_auth),
            ('presence_error', self.handler.on_presence_error),
            ('reactions', self.handler.on_reactions),
            ('receipt_received', self.handler.on_receipt),
            ('roster_subscription_authorized',
             self.handler.on_subscription_authorized),
            ('roster_subscription_remove',
             self.handler.on_subscription_remove),
            ('roster_subscription_removed',
             self.handler.on_subscription_removed),
            ('roster_subscription_request',
             self.handler.on_subscription_request),
            ('roster_update', self.handler.on_roster_update),
            ('session_start', self.handler.on_session_start),
            ('session_start', self.handler.on_session_start_features),
            ('session_end', self.handler.on_session_end),
            ('sm_failed', self.handler.on_session_end),
            ('session_resumed', self.handler.on_session_resumed),
            ('ssl_cert', self.handler.validate_ssl),
            ('ssl_invalid_chain', self.handler.ssl_invalid_chain),
            ('stream_error', self.handler.on_stream_error),
        ]
        for name, handler in xmpp_event_handlers:
            self.xmpp.add_event_handler(name, handler)

        if config.getbool('enable_avatars'):
            self.xmpp.add_event_handler("vcard_avatar_update",
                                        self.handler.on_vcard_avatar)
            self.xmpp.add_event_handler("avatar_metadata_publish",
                                        self.handler.on_0084_avatar)
        if config.getbool('enable_user_nick'):
            self.xmpp.add_event_handler("user_nick_publish",
                                        self.handler.on_nick_received)
        all_stanzas = Callback('custom matcher', connection.MatchAll(None),
                               self.handler.incoming_stanza)
        self.xmpp.register_handler(all_stanzas)

        self.initial_joins = []

        self.pending_invites = {}

        # a dict of the form {'config_option': [list, of, callbacks]}
        # Whenever a configuration option is changed (using /set or by
        # reloading a new config using a signal), all the associated
        # callbacks are triggered.
        # Use Core.add_configuration_handler("option", callback) to add a
        # handler
        # Note that the callback will be called when it’s changed in the
        # global section, OR in a special section.
        # As a special case, handlers can be associated with the empty
        # string option (""), they will be called for every option change
        # The callback takes two argument: the config option, and the new
        # value
        self.configuration_change_handlers = defaultdict(list)
        self.tab_configuration_change_handlers = defaultdict(list)
        config_handlers: List[Tuple[str, Callable[..., Any]]] = [
            ('', self.on_any_config_change),
            ('ack_message_receipts', self.on_ack_receipts_config_change),
            ('connection_check_interval', self.xmpp.set_keepalive_values),
            ('connection_timeout_delay', self.xmpp.set_keepalive_values),
            ('create_gaps', self.on_gaps_config_change),
            ('enable_carbons', self.on_carbons_switch),
            ('enable_reactions', self.on_reactions_change),
            ('enable_vertical_tab_list',
             self.on_vertical_tab_list_config_change),
            ('hide_user_list', self.on_hide_user_list_change),
            ('password', self.on_password_change),
            ('plugins_conf_dir',
             self.plugin_manager.on_plugins_conf_dir_change),
            ('plugins_dir', self.plugin_manager.on_plugins_dir_change),
            ('request_message_receipts',
             self.on_request_receipts_config_change),
            ('show_timestamps', self.on_show_timestamps_changed),
            ('theme', self.on_theme_config_change),
            ('themes_dir', theming.update_themes_dir),
            ('use_bookmarks_method', self.on_bookmarks_method_config_change),
            ('vertical_tab_list_size',
             self.on_vertical_tab_list_config_change),
            ('hide_exit_join', self.on_hide_exit_join_change),
        ]
        for option, handler in config_handlers:
            self.add_configuration_handler(option, handler)
        tab_config_handlers: List[Tuple[str, Callable[[str, str, str], Any]]] = [
            ('hide_exit_join', self.on_hide_exit_join_change),
        ]
        for option, handler in tab_config_handlers:
            self.add_tab_configuration_handler(option, handler)

    def _create_windows(self) -> None:
        """Create the windows (delayed after curses init)"""
        self.information_win = windows.TextWin(300)
        self.information_buffer.add_window(self.information_win)
        self.left_tab_win = None
        self.tab_win = windows.GlobalInfoBar(self)

    def on_tab_change(self, old_tab: Tab, new_tab: Tab) -> None:
        """Whenever the current tab changes, change focus and refresh"""
        old_tab.on_lose_focus()
        new_tab.on_gain_focus()
        self.refresh_window()

    def on_any_config_change(self, option: str, value: str) -> None:
        """
        Update the roster, in case a roster option changed.
        """
        roster.modified()

    def add_configuration_handler(self, option: str, callback: Callable[[str, str], None]) -> None:
        """
        Add a callback, associated with the given option. It will be called
        each time the configuration option is changed using /set or by
        reloading the configuration with a signal
        """
        self.configuration_change_handlers[option].append(callback)

    def add_tab_configuration_handler(self, option: str, callback: Callable[[str, str, str], None]) -> None:
        """
        Add a callback, associated with the given option. It will be called
        each time the configuration option is changed using /set or by
        reloading the configuration with a signal
        """
        self.tab_configuration_change_handlers[option].append(callback)

    def trigger_configuration_change(self, option: str, value: str, section: str | None = None) -> None:
        """
        Triggers all the handlers associated with the given configuration
        option
        """
        if section in (None, USE_DEFAULT_SECTION):
            # First call the callbacks associated with any configuration change
            for callback in self.configuration_change_handlers[""]:
                callback(option, value)
            # and then the callbacks associated with this specific option, if
            # any
            if option not in self.configuration_change_handlers:
                return
            for callback in self.configuration_change_handlers[option]:
                callback(option, value)
        else:
            for callback in self.tab_configuration_change_handlers[""]:
                callback(option, value, section)
            if option not in self.tab_configuration_change_handlers:
                return
            for callback in self.tab_configuration_change_handlers[option]:
                callback(option, value, section=section)

    def on_hide_user_list_change(self, option: str, value: str) -> None:
        """
        Called when the hide_user_list option changes
        """
        self.call_for_resize()

    def on_show_timestamps_changed(self, option: str, value: str) -> None:
        """
        Called when the show_timestamps option changes
        """
        self.call_for_resize(ui_config_changed=True)

    def on_reactions_change(self, option: str, value: str) -> None:
        """
        Called when the enable_reactions option changes
        """
        self.call_for_resize(ui_config_changed=True)

    def on_hide_exit_join_change(self, option: str, value: str, section: str = '') -> None:
        """
        Called when the hide_exit_join option changes.
        """
        if section:
            tab = self.tabs.by_name_and_class(section, MucTab)
            if tab is None:
                return
            tabs = [tab]
        else:
            tabs = self.tabs.by_class(MucTab)
            # Sucks a bit: we want to find out if the value of the option has
            # been changed for this specific tab, and remove it from the
            # affected tabs if that is the case.
            for tab in tabs[:]:
                custom_value = config.get_by_tabname(option, tab.jid, fallback=False)
                if custom_value != '' and custom_value != value:
                    tabs.remove(tab)
        for tab in tabs:
            tab.set_hide_exit_join(value)
        if isinstance(self.tabs.current_tab, MucTab):
            self.refresh_window()

    def on_bookmarks_method_config_change(self, option: str, value: str) -> None:
        """
        Called when the use_bookmarks_method option changes
        """
        if value not in ('pep', 'privatexml'):
            return
        self.bookmarks.preferred = value
        self.register_task(
            self.bookmarks.save(self.xmpp, core=self)
        )

    def on_gaps_config_change(self, option: str, value: str) -> None:
        """
        Called when the option create_gaps is changed.
        Remove all gaptabs if switching from gaps to nogaps.
        """
        self.tabs.update_gaps(value.lower() != "false")

    def on_request_receipts_config_change(self, option: str, value: str) -> None:
        """
        Called when the request_message_receipts option changes
        """
        self.xmpp.plugin['xep_0184'].auto_request = config.get(
            option, default=True)

    def on_ack_receipts_config_change(self, option: str, value: str) -> None:
        """
        Called when the ack_message_receipts option changes
        """
        self.xmpp.plugin['xep_0184'].auto_ack = config.get(
            option, default=True)

    def on_vertical_tab_list_config_change(self, option: str, value: str) -> None:
        """
        Called when the enable_vertical_tab_list option is changed
        """
        self.call_for_resize()

    def on_theme_config_change(self, option: str, value: str) -> None:
        """
        Called when the theme option is changed
        """
        error_msg = theming.reload_theme()
        if error_msg:
            self.information(error_msg, 'Warning')
        self.refresh_window()

    def on_password_change(self, option: str, value: str) -> None:
        """
        Set the new password in the slixmpp.ClientXMPP object
        """
        self.xmpp.password = value

    def on_carbons_switch(self, option: str, value: str) -> None:
        """Whenever the user enables or disables carbons using /set, we should
        inform the server immediately, this way we do not require a restart
        for the change to take effect
        """
        if value:
            self.xmpp.plugin['xep_0280'].enable()
        else:
            self.xmpp.plugin['xep_0280'].disable()

    def reload_config(self) -> None:
        # reload all log files
        log.debug("Reloading the log files…")
        logger.reload_all()
        log.debug("Log files reloaded.")
        # reload the theme
        log.debug("Reloading the theme…")
        theming.reload_theme()
        log.debug("Theme reloaded.")
        # reload the config from the disk
        log.debug("Reloading the config…")
        # Copy the old config in a dict
        old_config = config.to_dict()
        config.read_file()
        # Compare old and current config, to trigger the callbacks of all
        # modified options
        for section in config.sections():
            old_section = old_config.get(section, {})
            for option in config.options(section):
                old_value = old_section.get(option)
                new_value = config.get(option, default="", section=section)
                if new_value != old_value:
                    self.trigger_configuration_change(option, new_value)
        log.debug("Config reloaded.")
        for name, plugin in self.plugin_manager.plugins.items():
            if plugin:
                plugin.config.read_file()
                log.debug("Config reloaded for plugin %s", name)
        # in case some roster options have changed
        roster.modified()

    def sigusr_handler(self, num: Any, stack: Any) -> None:
        """
        Handle SIGUSR1 (10)
        When caught, reload all the possible files.
        """
        log.debug("SIGUSR1 caught, reloading the files…")
        self.reload_config()

    def exit_from_signal(self, *args: Any, **kwargs: Any) -> None:
        """
        Quit when receiving SIGHUP or SIGTERM or SIGPIPE

        do not save the config because it is not a normal exit
        (and only roster UI things are not yet saved)
        """
        sig = args[0]
        signals = {
            1: 'SIGHUP',
            13: 'SIGPIPE',
            15: 'SIGTERM',
        }

        log.error("%s received. Exiting…", signals[sig])
        self.plugin_manager.disable_plugins()
        self.disconnect('%s received' % signals.get(sig))
        self.xmpp.add_event_handler("disconnected", self.exit, disposable=True)

    def autoload_plugins(self) -> None:
        """
        Load the plugins on startup.
        """
        plugins = config.getstr('plugins_autoload')
        if ':' in plugins:
            for plugin in plugins.split(':'):
                self.plugin_manager.load(plugin, unload_first=False)
        else:
            for plugin in plugins.split():
                self.plugin_manager.load(plugin, unload_first=False)
        self.plugins_autoloaded = True

    def start(self) -> None:
        """
        Init curses, create the first tab, etc
        """
        self.stdscr = curses.initscr()
        self._init_curses(self.stdscr)
        windows.base_wins.TAB_WIN = self.stdscr
        self._create_windows()
        self.call_for_resize()
        default_tab = RosterInfoTab(self)
        default_tab.on_gain_focus()
        self.tabs.append(default_tab)
        self.information('Welcome to poezio!', 'Info')
        if curses.COLORS < 256:
            self.information(
                'Your terminal does not appear to support 256 colors, the UI'
                ' colors will probably be ugly',
                'Error',
            )
        if self.firstrun:
            self.information(
                'It seems that it is the first time you start poezio.\n'
                'The online help is here https://doc.poez.io/\n\n'
                'No room is joined by default, but you can join poezio’s'
                ' room (with \x19b/join poezio@muc.poez.io\x19o), where you can'
                ' ask for help or tell us how great it is.\n\n'
                'Note that all of your discussions are currently logged'
                ' to the disk, you can prevent that with'
                ' \x19b/set use_log false\x19o', 'Help')
        self.refresh_window()
        self.xmpp.plugin['xep_0012'].begin_idle(jid=self.xmpp.boundjid)

    def exit(self, event: str | None = None) -> None:
        log.debug("exit(%s)", event)
        asyncio.get_event_loop().stop()

    def on_exception(self, typ: Type[BaseException], value: BaseException,
                     trace: TracebackType | None) -> None:
        """
        When an exception is raised, just reset curses and call
        the original exception handler (will nicely print the traceback)
        """
        try:
            self.reset_curses()
        except Exception:
            log.debug('Unable to reset curses', exc_info=True)
        sys.__excepthook__(typ, value, trace)

    def sigwinch_handler(self, *args: Any) -> None:
        """A work-around for ncurses resize stuff, which sucks. Normally, ncurses
        catches SIGWINCH itself. In its signal handler, it updates the
        windows structures (for example the size, etc) and it
        ungetch(KEY_RESIZE). That way, the next time we call getch() we know
        that a resize occurred and we can act on it. BUT poezio doesn’t call
        getch() until it knows it will return something. The problem is we
        can’t know that, because stdin is not affected by this KEY_RESIZE
        value (it is only inserted in a ncurses internal fifo that we can’t
        access).

        The (ugly) solution is to handle SIGWINCH ourself, trigger the
        change of the internal windows sizes stored in ncurses module, using
        sizes that we get using shutil, ungetch the KEY_RESIZE value and
        then call getch to handle the resize on poezio’s side properly.
        """
        size = shutil.get_terminal_size()
        curses.resizeterm(size.lines, size.columns)
        curses.ungetch(curses.KEY_RESIZE)
        self.on_input_readable()

    def on_input_readable(self) -> None:
        """
        main loop waiting for the user to press a key
        """

        log.debug("Input is readable.")
        big_char_list = [replace_key_with_bound(key)
                         for key in self.read_keyboard()]
        log.debug("Got from keyboard: %s", (big_char_list, ))

        # whether to refresh after ALL keys have been handled
        for char_list in separate_chars_from_bindings(big_char_list):
            # Special case for M-x where x is a number
            if len(char_list) == 1:
                char = char_list[0]
                if char.startswith('M-') and len(char) == 3:
                    try:
                        nb = int(char[2])
                    except ValueError:
                        pass
                    else:
                        if self.tabs.current_tab.nb == nb and config.getbool(
                                'go_to_previous_tab_on_alt_number'):
                            self.go_to_previous_tab()
                        else:
                            self.command.win('%d' % nb)
                # search for keyboard shortcut
                func = self.key_func.get_cb(char, None)
                if func is not None:
                    func()
                else:
                    self.do_command(replace_line_breaks(char), False)
            else:
                self.do_command(''.join(char_list), True)
        self.doupdate()

    def loop_exception_handler(self, loop: AbstractEventLoop, context: dict[str, Any]) -> None:
        """Do not log unhandled iq errors and timeouts"""
        handled_exceptions = (IqError, IqTimeout, InvalidCABundle)
        # Can happen when the loop is being shut down
        if 'exception' not in context:
            if 'message' in context:
                log.error('Unhandled error: %s', context['message'])
        elif not isinstance(context['exception'], handled_exceptions):
            loop.default_exception_handler(context)
        elif isinstance(context['exception'], InvalidCABundle):
            paths = context['exception'].path
            error = (
                'Poezio could not find a valid CA bundle file automatically. '
                'Ensure the ca_cert_path configuration is set to a valid '
                'CA bundle path, generally provided by the \'ca-certificates\' '
                'package in your distribution.'
            )
            if isinstance(paths, (str, Path)):
                # error += '\nFound the following value: {path}'.format(path=str(path))
                paths = [paths]
            if paths is not None:
                error += f"\nThe following values were tried: {str([str(s) for s in paths])}"
            self.information(error, 'Error')

    def save_config(self) -> None:
        """
        Save config in the file just before exit
        """
        ok = roster.save_to_config_file()
        ok = ok and config.silent_set('info_win_height',
                                      self.information_win_size, 'var')
        if not ok:
            self.information(
                'Unable to save runtime preferences'
                ' in the config file', 'Error')

    def on_roster_enter_key(self, roster_row: Contact | Resource | RosterGroup) -> None:
        """
        when enter is pressed on the roster window
        """
        if isinstance(roster_row, Contact):
            if not self.get_conversation_by_jid(roster_row.bare_jid, False):
                self.open_conversation_window(JID(roster_row.bare_jid))
            else:
                self.focus_tab_named(roster_row.bare_jid.bare)
        if isinstance(roster_row, Resource):
            if not self.get_conversation_by_jid(
                    roster_row.jid, False, fallback_barejid=False):
                self.open_conversation_window(JID(roster_row.jid))
            else:
                self.focus_tab_named(roster_row.jid)
        self.refresh_window()

    def get_conversation_messages(self) -> list[BaseMessage]:
        """
        Returns a list of all the messages in the current chat.
        If the current tab is not a ChatTab, returns None.

        Messages are namedtuples of the form
        ('txt nick_color time str_time nickname user')
        """
        if not isinstance(self.tabs.current_tab, ChatTab):
            return []
        return self.tabs.current_tab.get_conversation_messages()

    def insert_input_text(self, text: str) -> None:
        """
        Insert the given text into the current input
        """
        self.do_command(text, True)

##################### Anything related to command execution ###################

    def execute(self, line: str) -> None:
        """
        Execute the /command or just send the line on the current room
        """
        if line == "":
            return
        if line.startswith('/'):
            command = line.strip().split()[0][1:]
            arg = line[2 + len(command):]  # jump the '/' and the ' '
            # example. on "/link 0 open", command = "link" and arg = "0 open"
            if command in self.commands:
                func = self.commands[command].func
                func(arg)
                return
            else:
                self.information("Unknown command (%s)" % (command), 'Error')

    def exec_command(self, command: list[str]) -> None:
        """
        Execute an external command on the local or a remote machine,
        depending on the conf. For example, to open a link in a browser, do
        exec_command(["firefox", "http://poezio.eu"]), and this will call
        the command on the correct computer.

        The command argument is a list of strings, not quoted or escaped in
        any way. The escaping is done here if needed.

        The remote execution is done
        by writing the command on a fifo.  That fifo has to be on the
        machine where poezio is running, and accessible (through sshfs for
        example) from the local machine (where poezio is not running). A
        very simple daemon (daemon.py) reads on that fifo, and executes any
        command that is read in it. Since we can only write strings to that
        fifo, each argument has to be shlex.quote()d. That way the
        shlex.split on the reading-side of the daemon will be safe.

        You cannot use a real command line with pipes, redirections etc, but
        this function supports a simple case redirection to file: if the
        before-last argument of the command is ">" or ">>", then the last
        argument is considered to be a filename where the command stdout
        will be written. For example you can do exec_command(["echo",
        "coucou les amis coucou coucou", ">", "output.txt"]) and this will
        work. If you try to do anything else, your |, [, <<, etc will be
        interpreted as normal command arguments, not shell special tokens.
        """
        if config.getbool('exec_remote'):
            # We just write the command in the fifo
            fifo_path = config.getstr('remote_fifo_path')
            filename = os.path.join(fifo_path, 'poezio.fifo')
            if not self.remote_fifo:
                try:
                    self.remote_fifo = Fifo(filename, 'w')
                except OSError as exc:
                    log.exception(
                        'Could not open the fifo for writing (%s)',
                        filename,
                    )
                    self.information(
                        'Could not open the fifo '
                        'file for writing: %s' % exc, 'Error')
                    return

            args = (shlex.quote(arg.replace('\n', ' ')) for arg in command)
            command_str = ' '.join(args) + '\n'
            try:
                self.remote_fifo.write(command_str)
            except IOError as exc:
                log.exception(
                    'Could not write in the fifo (%s): %s',
                    filename,
                    repr(command),
                )
                self.information('Could not execute %s: %s' % (command, exc),
                                 'Error')
                self.remote_fifo = None
        else:
            executor = Executor(command)
            try:
                executor.start()
            except ValueError as exc:
                log.exception(
                    'Could not execute command (%s)',
                    repr(command),
                )
                self.information(str(exc), 'Error')

    def do_command(self, key: str, raw: bool) -> None:
        """
        Execute the action associated with a key

        Or if keyboard.continuation_keys_callback is set, call it instead. See
        the comment of this variable.
        """
        if not key:
            return
        if keyboard.continuation_keys_callback is not None:
            # Reset the callback to None BEFORE calling it, because this
            # callback MAY set a new callback itself, and we don’t want to
            # erase it in that case
            cb = keyboard.continuation_keys_callback
            keyboard.continuation_keys_callback = None
            cb(key)
        else:
            self.tabs.current_tab.on_input(key, raw)

    def try_execute(self, line: str) -> None:
        """
        Try to execute a command in the current tab
        """
        line = '/' + line
        try:
            self.tabs.current_tab.execute_command(line)
        except Exception:
            log.exception('Execute failed (%s)', line)

########################## TImed Events #######################################

    def remove_timed_event(self, event: DelayedEvent) -> None:
        """Remove an existing timed event"""
        if event.handler is not None:
            event.handler.cancel()

    def add_timed_event(self, event: DelayedEvent) -> None:
        """Add a new timed event"""
        event.handler = asyncio.get_event_loop().call_later(
            event.delay, event.callback, *event.args
        )

####################### XMPP-related actions ##################################

    def get_status(self) -> Status:
        """
        Get the last status that was previously set
        """
        return self.status

    def set_status(self, pres: str, msg: str) -> None:
        """
        Set our current status so we can remember
        it and use it back when needed (for example to display it
        or to use it when joining a new muc)
        """
        self.status = Status(show=pres, message=msg)
        if config.getbool('save_status'):
            ok = config.silent_set('status', pres if pres else '')
            msg = msg.replace('\n', '|') if msg else ''
            ok = ok and config.silent_set('status_message', msg)
            if not ok:
                self.information(
                    'Unable to save the status in '
                    'the config file', 'Error')

    def get_bookmark_nickname(self, room_name: str) -> str:
        """
        Returns the nickname associated with a bookmark
        or the default nickname
        """
        bm = self.bookmarks[room_name]
        if bm and bm.nick:
            return bm.nick
        return self.own_nick

    def disconnect(self, msg: str = '', reconnect: bool = False) -> None:
        """
        Disconnect from remote server and correctly set the states of all
        parts of the client (for example, set the MucTabs as not joined, etc)
        """
        self.legitimate_disconnect = True
        if reconnect:
            self.xmpp.reconnect(wait=0.0, reason=msg)
        else:
            for tab in self.get_tabs(MucTab):
                tab.leave_room(msg)
            self.xmpp.disconnect(reason=msg)

    def send_message(self, msg: str) -> bool:
        """
        Function to use in plugins to send a message in the current
        conversation.
        Returns False if the current tab is not a conversation tab
        """
        if not isinstance(self.tabs.current_tab, ChatTab):
            return False
        asyncio.ensure_future(
            self.tabs.current_tab.command_say(msg)
        )
        return True

    async def invite(self, jid: JID, room: JID, reason: Optional[str] = None,
                     force_mediated: bool = False) -> bool:
        """
        Checks if the sender supports XEP-0249, then send an invitation,
        or a mediated one if it does not.
        TODO: allow passwords
        """
        features = set()

        # force mediated: act as if the other entity does not
        # support direct invites
        if not force_mediated:
            try:
                iq = await self.xmpp.plugin['xep_0030'].get_info(
                    jid=jid,
                    timeout=5,
                )
                features = iq['disco_info'].get_features()
            except (IqError, IqTimeout):
                pass
        supports_direct = 'jabber:x:conference' in features
        if supports_direct:
            self.xmpp.plugin['xep_0249'].send_invitation(
                jid=jid,
                roomjid=room,
                reason=reason
            )
        else:  # fallback
            self.xmpp.plugin['xep_0045'].invite(
                jid=jid,
                room=room,
                reason=reason or '',
            )
        return True

    def _impromptu_room_form(self, room: JID) -> Iq:
        fields = [
            ('hidden', 'FORM_TYPE', 'http://jabber.org/protocol/muc#roomconfig'),
            ('boolean', 'muc#roomconfig_changesubject', True),
            ('boolean', 'muc#roomconfig_allowinvites', True),
            ('boolean', 'muc#roomconfig_persistent', True),
            ('boolean', 'muc#roomconfig_membersonly', True),
            ('boolean', 'muc#roomconfig_publicroom', False),
            ('list-single', 'muc#roomconfig_whois', 'anyone'),
            # MAM
            ('boolean', 'muc#roomconfig_enablearchiving', True),  # Prosody
            ('boolean', 'mam', True),  # Ejabberd community
            ('boolean', 'muc#roomconfig_mam', True),  # Ejabberd saas
        ]

        form = self.xmpp['xep_0004'].make_form()
        form['type'] = 'submit'
        for field in fields:
            form.add_field(
                ftype=field[0],
                var=field[1],
                value=field[2],
            )

        iq = self.xmpp.Iq()
        iq['type'] = 'set'
        iq['to'] = room
        query = ET.Element('{http://jabber.org/protocol/muc#owner}query')
        query.append(form.xml)
        iq.append(query)
        return iq

    async def impromptu(self, jids: Set[JID]) -> None:
        """
        Generates a new "Impromptu" room with a random localpart on the muc
        component of the user who initiated the request. One the room is
        created and the first user has joined, send invites for specified
        contacts to join in.
        """

        results = await self.xmpp['xep_0030'].get_info_from_domain()

        muc_from_identity = ''
        for info in results:
            for identity in info['disco_info']['identities']:
                if identity[0] == 'conference' and identity[1] == 'text':
                    muc_from_identity = info['from'].bare

        # Use config.default_muc_service as muc component if available,
        # otherwise find muc component by disco#items-ing the user domain.
        # If not, give up
        default_muc = config.get('default_muc_service', muc_from_identity)
        if not default_muc:
            self.information(
                "Error finding a MUC service to join. If your server does not "
                "provide one, set 'default_muc_service' manually to a MUC "
                "service that allows room creation.",
                'Error'
            )
            return

        # Retries generating a name until we find a non-existing room.
        # Abort otherwise.
        retries = 3
        while retries > 0:
            localpart = utils.pronounceable()
            room_str = f'{localpart}@{default_muc}'
            try:
                room = JID(room_str)
            except InvalidJID:
                self.information(
                    f'The generated XMPP address is invalid: {room_str}',
                    'Error'
                )
                return None

            try:
                await self.xmpp['xep_0030'].get_info(
                    jid=room,
                    cached=False,
                )
            except IqTimeout:
                pass
            except IqError as exn:
                if exn.etype == 'cancel' and exn.condition == 'item-not-found':
                    log.debug('Found empty room for /impromptu')
                    break

            retries = retries - 1

        if retries == 0:
            self.information(
                'Couldn\'t generate a room name that isn\'t already used.',
                'Error',
            )
            return None

        self.open_new_room(room, self.own_nick).join()

        async def configure_and_invite(_presence: Presence) -> None:
            iq = self._impromptu_room_form(room)
            try:
                await iq.send()
            except (IqError, IqTimeout):
                self.information('Failed to configure impromptu room.', 'Info')
                # TODO: destroy? leave room.
                return

            self.information(f'Room {room} created', 'Info')

            for jid in jids:
                await self.invite(jid, room, force_mediated=True)
            jids_str = ', '.join(map(lambda j: j.full, jids))
            self.information(f'Invited {jids_str} to {room.bare}', 'Info')

        self.xmpp.add_event_handler(
            f'muc::{room.bare}::groupchat_subject',
            configure_and_invite,
            disposable=True,
        )

####################### Tab logic-related things ##############################

### Tab getters ###

    def get_tabs(self, cls: Type[T]) -> List[T]:
        "Get all the tabs of a type"
        return self.tabs.by_class(cls)

    def get_conversation_by_jid(self,
                                jid: JID | str,
                                create: bool = True,
                                fallback_barejid: bool = True) -> ConversationTab | None:
        """
        From a JID, get the tab containing the conversation with it.
        If none already exist, and create is "True", we create it
        and return it. Otherwise, we return None.

        If fallback_barejid is True, then this method will seek other
        tabs with the same barejid, instead of searching only by fulljid.
        """
        jid = JID(jid)
        # We first check if we have a static conversation opened
        # with this precise resource
        conversation: Optional[ConversationTab]
        conversation = self.tabs.by_name_and_class(jid.full,
                                                   StaticConversationTab)
        if jid.bare == jid.full and not conversation:
            conversation = self.tabs.by_name_and_class(
                jid.full, DynamicConversationTab)

        if not conversation and fallback_barejid:
            # If not, we search for a conversation with the bare jid
            conversation = self.tabs.by_name_and_class(
                jid.bare, DynamicConversationTab)
            if not conversation:
                if create:
                    # We create a dynamic conversation with the bare Jid if
                    # nothing was found (and we lock it to the resource
                    # later)
                    conversation = self.open_conversation_window(
                        JID(jid.bare), False)
                else:
                    conversation = None
        return conversation

    def add_tab(self, new_tab: Tab, focus: bool = False) -> None:
        """
        Appends the new_tab in the tab list and
        focus it if focus==True
        """
        self.tabs.append(new_tab)
        if focus:
            self.tabs.set_current_tab(new_tab)

    def insert_tab(self, old_pos: int, new_pos: int = 99999) -> bool:
        """
        Insert a tab at a position, changing the number of the following tabs
        returns False if it could not move the tab, True otherwise
        """
        return self.tabs.insert_tab(old_pos, new_pos,
                                    config.getbool('create_gaps'))

    ### Move actions (e.g. go to next room) ###

    def rotate_rooms_right(self) -> None:
        """
        rotate the rooms list to the right
        """
        self.tabs.next()  # pylint: disable=not-callable

    def rotate_rooms_left(self) -> None:
        """
        rotate the rooms list to the right
        """
        self.tabs.prev()  # pylint: disable=not-callable

    def go_to_room_number(self) -> None:
        """
        Read 2 more chars and go to the tab
        with the given number
        """

        def read_next_digit(digit: str) -> None:
            try:
                int(digit)
            except ValueError:
                # If it is not a number, we do nothing. If it was the first
                # one, we do not wait for a second one by re-setting the
                # callback
                self.room_number_jump.clear()
            else:
                self.room_number_jump.append(digit)
                if len(self.room_number_jump) == 2:
                    arg = "".join(self.room_number_jump)
                    self.room_number_jump.clear()
                    self.command.win(arg)
                else:
                    # We need to read more digits
                    keyboard.continuation_keys_callback = read_next_digit

        keyboard.continuation_keys_callback = read_next_digit

    def go_to_room_name(self) -> None:
        room_name_jump = []

        def read_next_letter(s: str) -> None:
            room_name_jump.append(s)
            any_matched, unique_tab = self.tabs.find_by_unique_prefix(
                "".join(room_name_jump)
            )

            if not any_matched:
                return

            if unique_tab is not None:
                self.tabs.set_current_tab(unique_tab)
                # NOTE: returning here means that as soon as the tab is
                # matched, normal input resumes. If we do *not* return here,
                # any further characters matching the prefix of the tab will
                # be swallowed (and a lot of tab switching will happen...),
                # until a non-matching character or escape or something is
                # pressed.
                # This behaviour *may* be desirable.
                return

            keyboard.continuation_keys_callback = read_next_letter

        keyboard.continuation_keys_callback = read_next_letter

    def go_to_roster(self) -> None:
        "Select the roster as the current tab"
        self.tabs.set_current_tab(self.tabs.first())

    def go_to_previous_tab(self) -> None:
        "Go to the previous tab"
        self.tabs.restore_previous_tab()

    def go_to_important_room(self) -> None:
        """
        Go to the next room with activity, in the order defined in the
        dict STATE_PRIORITY
        """
        # shortcut
        priority = STATE_PRIORITY
        tab_refs: Dict[str, List[Tab]] = {}
        # put all the active tabs in a dict of lists by state
        for tab in self.tabs.get_tabs():
            if not tab:
                continue
            if tab.state not in tab_refs:
                tab_refs[tab.state] = [tab]
            else:
                tab_refs[tab.state].append(tab)
        # sort the state by priority and remove those with negative priority
        states = sorted(
            tab_refs.keys(), key=(lambda x: priority.get(x, 0)), reverse=True)
        states = [state for state in states if priority.get(state, -1) >= 0]

        for state in states:
            for tab in tab_refs[state]:
                if (tab.nb < self.tabs.current_index
                        and tab_refs[state][-1].nb > self.tabs.current_index):
                    continue
                self.tabs.set_current_tab(tab)
                return
        return

    def focus_tab_named(self,
                        tab_name: str,
                        type_: Optional[Type[Tab]] = None) -> bool:
        """Returns True if it found a tab to focus on"""
        if type_ is None:
            tab = self.tabs.by_name(tab_name)
        else:
            tab = self.tabs.by_name_and_class(tab_name, type_)
        if tab:
            self.tabs.set_current_tab(tab)
            return True
        return False

    def focus_tab(self, tab: Tab) -> bool:
        """Focus a tab"""
        return self.tabs.set_current_tab(tab)

    ### Opening actions ###

    def open_conversation_window(self, jid: JID,
                                 focus: bool = True) -> ConversationTab:
        """
        Open a new conversation tab and focus it if needed. If a resource is
        provided, we open a StaticConversationTab, else a
        DynamicConversationTab
        """
        new_tab: ConversationTab
        if jid.resource:
            new_tab = StaticConversationTab(self, jid)
        else:
            new_tab = DynamicConversationTab(self, jid)
        if not focus:
            new_tab.state = "private"
        self.add_tab(new_tab, focus)
        self.refresh_window()
        return new_tab

    def open_private_window(self, room_name: str, user_nick: str,
                            focus: bool = True) -> Optional[PrivateTab]:
        """
        Open a Private conversation in a MUC and focus if needed.
        """
        complete_jid = room_name + '/' + user_nick
        # if the room exists, focus it and return
        for tab in self.get_tabs(PrivateTab):
            if tab.name == complete_jid:
                self.tabs.set_current_tab(tab)
                return tab
        # create the new tab
        muc_tab = self.tabs.by_name_and_class(room_name, MucTab)
        if muc_tab is None:
            return None
        tab = PrivateTab(self, JID(complete_jid), muc_tab.own_nick, muc_tab)
        if hasattr(tab, 'directed_presence'):
            tab.directed_presence = tab.directed_presence
        if not focus:
            tab.state = "private"
        # insert it in the tabs
        self.add_tab(tab, focus)
        self.refresh_window()
        muc_tab.privates.append(tab)
        return tab

    def open_new_room(self,
                      room: JID,
                      nick: str,
                      *,
                      password: Optional[str] = None,
                      focus: bool = True) -> MucTab:
        """
        Open a new tab.MucTab containing a muc Room, using the specified nick
        """
        new_tab = MucTab(self, room, nick, password=password)
        self.add_tab(new_tab, focus)
        self.refresh_window()
        return new_tab

    def open_new_form(self, form: Form, on_cancel: Callable[Concatenate[Form, P], None], on_send: Callable[Concatenate[Form, P], None],
                      *args: P.args, **kwargs: P.kwargs) -> None:
        """
        Open a new tab containing the form
        The callback are called with the completed form as parameter in
        addition with kwargs
        """
        form_tab = DataFormsTab(self, form, on_cancel, on_send, *args, **kwargs)
        self.add_tab(form_tab, True)

    ### Modifying actions ###

    def rename_private_tabs(self, room_name: str, old_nick: str, user: User) -> None:
        """
        Call this method when someone changes their nick in a MUC,
        this updates the name of all the opened private conversations
        with him/her
        """
        tab = self.tabs.by_name_and_class('%s/%s' % (room_name, old_nick),
                                          PrivateTab)
        if tab:
            tab.rename_user(old_nick, user)

    def on_user_left_private_conversation(self, room_name: str, user: User,
                                          status_message: str) -> None:
        """
        The user left the MUC: add a message in the associated
        private conversation
        """
        tab = self.tabs.by_name_and_class('%s/%s' % (room_name, user.nick),
                                          PrivateTab)
        if tab:
            tab.user_left(status_message, user)

    def on_user_rejoined_private_conversation(self, room_name: str, nick: str) -> None:
        """
        The user joined a MUC: add a message in the associated
        private conversation
        """
        tab = self.tabs.by_name_and_class('%s/%s' % (room_name, nick),
                                          PrivateTab)
        if tab:
            tab.user_rejoined(nick)

    def disable_private_tabs(self,
                             room_name: str,
                             reason: Optional[str] = None) -> None:
        """
        Disable private tabs when leaving a room
        """
        if reason is None:
            reason = '\x195}You left the room\x193}'
        for tab in self.get_tabs(PrivateTab):
            if tab.name.startswith(room_name):
                tab.deactivate(reason=reason)

    def enable_private_tabs(self, room_name: str,
                            reason: Optional[str] = None) -> None:
        """
        Enable private tabs when joining a room
        """
        if reason is None:
            reason = '\x195}You joined the room\x193}'
        for tab in self.get_tabs(PrivateTab):
            if tab.name.startswith(room_name):
                tab.activate(reason=reason)

    def on_user_changed_status_in_private(self, jid: JID, status: Status) -> None:
        tab = self.tabs.by_name_and_class(jid, OneToOneTab)
        if tab is not None:  # display the message in private
            tab.update_status(status)

    def close_tab(self, to_close: Tab | None = None) -> None:
        """
        Close the given tab. If None, close the current one
        """
        was_current = to_close is None
        tab = to_close or self.tabs.current_tab

        if isinstance(tab, RosterInfoTab):
            return  # The tab 0 should NEVER be closed
        tab.on_close()
        del tab.key_func  # Remove self references
        del tab.commands  # and make the object collectable
        self.tabs.delete(tab, gap=config.getbool('create_gaps'))
        logger.close(tab.name)
        if was_current:
            self.tabs.current_tab.on_gain_focus()
        self.refresh_window()
        import gc
        gc.collect()
        log.debug('___ Referrers of closing tab:\n%s\n______',
                  gc.get_referrers(tab))
        del tab

    def add_information_message_to_conversation_tab(self, jid: JID, msg: str) -> None:
        """
        Search for a ConversationTab with the given jid (full or bare),
        if yes, add the given message to it
        """
        tab = self.tabs.by_name_and_class(jid, ConversationTab)
        if tab is not None:
            tab.add_message(PersistentInfoMessage(msg))
            if self.tabs.current_tab is tab:
                self.refresh_window()

####################### Curses and ui-related stuff ###########################

    def doupdate(self) -> None:
        "Do a curses update"
        curses.doupdate()

    def information(self, msg: str, typ: str = '') -> bool:
        """
        Displays an informational message in the "Info" buffer
        """
        filter_types = config.getlist('information_buffer_type_filter')
        if typ.lower() in filter_types:
            log.debug(
                'Did not show the message:\n\t%s> %s \n\tdue to '
                'information_buffer_type_filter configuration', typ, msg)
            return False
        filter_messages = config.getlist('filter_info_messages')
        for words in filter_messages:
            if words and words in msg:
                log.debug(
                    (
                        'Did not show the message:\n\t%s> %s \n\t'
                        'due to filter_info_messages configuration'
                    ),
                    typ, msg)
                return False
        nb_lines = self.information_buffer.add_message(
            UIMessage(
                txt=msg,
                level=typ,
            )
        )
        popup_on = config.getlist('information_buffer_popup_on')
        if isinstance(self.tabs.current_tab, RosterInfoTab):
            self.refresh_window()
        elif typ != '' and typ.lower() in popup_on:
            popup_time = config.getint('popup_time') + (nb_lines - 1) * 2
            self._pop_information_win_up(nb_lines, popup_time)
        else:
            if self.information_win_size != 0:
                self.information_win.refresh()
                self.tabs.current_tab.refresh_input()
        return True

    def _init_curses(self, stdscr: curses.window) -> None:
        """
        ncurses initialization
        """
        curses.curs_set(1)
        curses.noecho()
        curses.nonl()
        curses.raw()
        stdscr.idlok(True)
        stdscr.keypad(True)
        curses.start_color()
        curses.use_default_colors()
        theming.reload_theme()
        curses.ungetch(" ")  # H4X: without this, the screen is
        stdscr.getkey()  # erased on the first "getkey()"

    def reset_curses(self) -> None:
        """
        Reset terminal capabilities to what they were before ncurses
        init
        """
        curses.echo()
        curses.nocbreak()
        curses.curs_set(1)
        curses.endwin()

    def refresh_window(self) -> None:
        """
        Refresh everything
        """
        nocursor = curses.curs_set(0)
        self.tabs.current_tab.state = 'current'
        self.tabs.current_tab.refresh()
        self.doupdate()
        curses.curs_set(nocursor)

    def refresh_tab_win(self) -> None:
        """
        Refresh the window containing the tab list
        """
        self.tabs.current_tab.refresh_tab_win()
        self.refresh_input()
        self.doupdate()

    def refresh_input(self) -> None:
        """
        Refresh the input if it exists
        """
        if self.tabs.current_tab.input:
            self.tabs.current_tab.input.refresh()
        self.doupdate()

    def scroll_page_down(self) -> bool:
        """
        Scroll a page down, if possible.
        Returns True on success, None on failure.
        """
        if self.tabs.current_tab.on_scroll_down():
            self.refresh_window()
            return True
        return False

    def scroll_page_up(self) -> bool:
        """
        Scroll a page up, if possible.
        Returns True on success, None on failure.
        """
        if self.tabs.current_tab.on_scroll_up():
            self.refresh_window()
            return True
        return False

    def scroll_line_up(self) -> bool:
        """
        Scroll a line up, if possible.
        Returns True on success, None on failure.
        """
        if self.tabs.current_tab.on_line_up():
            self.refresh_window()
            return True
        return False

    def scroll_line_down(self) -> bool:
        """
        Scroll a line down, if possible.
        Returns True on success, None on failure.
        """
        if self.tabs.current_tab.on_line_down():
            self.refresh_window()
            return True
        return False

    def scroll_half_up(self) -> bool:
        """
        Scroll half a screen down, if possible.
        Returns True on success, None on failure.
        """
        if self.tabs.current_tab.on_half_scroll_up():
            self.refresh_window()
            return True
        return False

    def scroll_half_down(self) -> bool:
        """
        Scroll half a screen down, if possible.
        Returns True on success, None on failure.
        """
        if self.tabs.current_tab.on_half_scroll_down():
            self.refresh_window()
            return True
        return False

    def grow_information_win(self, nb: int = 1) -> int:
        """
        Expand the information win a number of lines
        """
        if self.information_win_size >= self.tabs.current_tab.height - 5 or \
                self.information_win_size + nb >= self.tabs.current_tab.height - 4 or \
                self.size.core_degrade_y:
            return 0
        self.information_win_size += nb
        self.resize_global_information_win()
        for tab in self.tabs:
            tab.on_info_win_size_changed()
        self.refresh_window()
        return nb

    def shrink_information_win(self, nb: int = 1) -> None:
        """
        Reduce the size of the information win
        """
        if self.information_win_size == 0 or self.size.core_degrade_y:
            return
        self.information_win_size -= nb
        if self.information_win_size < 0:
            self.information_win_size = 0
        self.resize_global_information_win()
        for tab in self.tabs:
            tab.on_info_win_size_changed()
        self.refresh_window()

    def scroll_info_up(self) -> None:
        """
        Scroll the information buffer up
        """
        self.information_win.scroll_up(self.information_win.height)
        if not isinstance(self.tabs.current_tab, RosterInfoTab):
            self.information_win.refresh()
        else:
            info = self.tabs.current_tab.information_win
            info.scroll_up(info.height)
            self.refresh_window()

    def scroll_info_down(self) -> None:
        """
        Scroll the information buffer down
        """
        self.information_win.scroll_down(self.information_win.height)
        if not isinstance(self.tabs.current_tab, RosterInfoTab):
            self.information_win.refresh()
        else:
            info = self.tabs.current_tab.information_win
            info.scroll_down(info.height)
            self.refresh_window()

    def _pop_information_win_up(self, size: int, time: int) -> None:
        """
        Temporarily increase the size of the information win of size lines
        during time seconds.
        After that delay, the size will decrease from size lines.
        """
        if time <= 0 or size <= 0:
            return
        result = self.grow_information_win(size)
        timed_event = timed_events.DelayedEvent(
            time, self.shrink_information_win, result)
        self.add_timed_event(timed_event)
        self.refresh_window()

    def toggle_left_pane(self) -> None:
        """
        Enable/disable the left panel.
        """
        enabled = config.getbool('enable_vertical_tab_list')
        if not config.silent_set('enable_vertical_tab_list', str(not enabled)):
            self.information('Unable to write in the config file', 'Error')
        self.call_for_resize()

    def toggle_show_presences(self, local: bool = False) -> None:
        """
        Enable/disable showing presences in MUCs.
        """
        if not local:
            section = USE_DEFAULT_SECTION
            hide_exit_join = 0 if config.getint('hide_exit_join') != 0 else -1
        elif isinstance(self.tabs.current_tab, MucTab):
            section = str(self.tabs.current_tab.jid)
            hide_exit_join = 0 if config.getint('hide_exit_join', section=section) != 0 else -1
        else:
            return
        if not config.silent_set('hide_exit_join', str(hide_exit_join), section=section):
            self.information('Unable to write in the config file', 'Error')
        self.trigger_configuration_change('hide_exit_join', str(hide_exit_join), section=section)

    def resize_global_information_win(self, ui_config_changed: bool = False) -> None:
        """
        Resize the global_information_win only once at each resize.
        """
        if self.information_win_size > Tab.height - 6:
            self.information_win_size = Tab.height - 6
        if Tab.height < 6:
            self.information_win_size = 0
        height = (Tab.height - 1 - self.information_win_size -
                  Tab.tab_win_height())
        self.information_win.resize(self.information_win_size, Tab.width,
                                    height, 0, self.information_buffer,
                                    force=ui_config_changed)

    def resize_global_info_bar(self) -> None:
        """
        Resize the GlobalInfoBar only once at each resize
        """
        if config.getbool('enable_vertical_tab_list'):

            if self.size.core_degrade_x:
                return
            if self.stdscr is None:
                return
            try:
                height, _ = self.stdscr.getmaxyx()
                truncated_win = self.stdscr.subwin(
                    height, config.getint('vertical_tab_list_size'), 0, 0)
            except Exception:
                log.exception('Curses error on infobar resize')
                return
            self.left_tab_win = windows.VerticalGlobalInfoBar(
                self, truncated_win)
        elif not self.size.core_degrade_y:
            self.tab_win.resize(1, Tab.width, Tab.height - 2, 0)
            self.left_tab_win = None

    def full_screen_redraw(self) -> None:
        """
        Completely erase and redraw the screen
        """
        if self.stdscr is not None:
            self.stdscr.clear()
            self.refresh_window()

    def call_for_resize(self, ui_config_changed: bool = False) -> None:
        """
        Called when we want to resize the screen
        """
        # If we have the tabs list on the left, we just give a truncated
        # window to each Tab class, so they draw themself in the portion of
        # the screen that they can occupy, and we draw the tab list on the
        # remaining space, on the left
        if self.stdscr is None:
            raise ScreenNotAvailable()
        height, width = self.stdscr.getmaxyx()
        if (config.getbool('enable_vertical_tab_list')
                and not self.size.core_degrade_x):
            try:
                scr = self.stdscr.subwin(
                    0,
                    config.getint('vertical_tab_list_size')
                )
            except Exception:
                log.exception('Curses error on resize')
                return
        else:
            scr = self.stdscr
        Tab.initial_resize(scr)
        self.resize_global_info_bar()
        self.resize_global_information_win(ui_config_changed)
        for tab in self.tabs:
            tab.ui_config_changed = True
            if config.getbool('lazy_resize'):
                tab.need_resize = True
            else:
                tab.resize()

        if len(self.tabs):
            self.full_screen_redraw()

    def read_keyboard(self) -> list[str]:
        """
        Get the next keyboard key pressed and returns it.  It blocks until
        something can be read on stdin, this function must be called only if
        there is something to read. No timeout ever occurs.
        """
        if self.stdscr is None:
            return []
        return self.keyboard.get_user_input(self.stdscr)

    def escape_next_key(self) -> None:
        """
        Tell the Keyboard object that the next key pressed by the user
        should be escaped. See Keyboard.get_user_input
        """
        self.keyboard.escape_next_key()

####################### Commands and completions ##############################

    def register_command(self, name: str, func: Callable[Concatenate[str, P], Any],
                         **kwargs: Unpack[CommandKwargs]) -> None:
        """
        Add a command
        """
        desc = kwargs.get('desc', '')
        shortdesc = kwargs.get('shortdesc', '')
        completion = kwargs.get('completion')
        usage = kwargs.get('usage', '')
        if name in self.commands:
            return
        if not desc and shortdesc:
            desc = shortdesc
        self.commands[name] = Command(func, desc, completion, shortdesc, usage)

    def register_initial_commands(self) -> None:
        """
        Register the commands when poezio starts
        """
        for command in get_commands(self.command, self.completion, self.plugin_manager):
            self.register_command(**command)

    def check_blocking(self, features: list[str]) -> None:
        """Check if the server supports the blocking feature."""
        if 'urn:xmpp:blocking' in features and not self.xmpp.anon:
            self.register_command(
                'block',
                self.command.block,
                usage='[jid]',
                shortdesc='Prevent a JID from talking to you.',
                completion=self.completion.block)
            self.register_command(
                'unblock',
                self.command.unblock,
                usage='[jid]',
                shortdesc='Allow a JID to talk to you.',
                completion=self.completion.unblock)

            self.xmpp.del_event_handler('session_start', self.check_blocking)

####################### Random things to move #################################

    def join_initial_rooms(self, bookmarks: list[Bookmark]) -> None:
        """Join all rooms given in the iterator `bookmarks`"""
        for bm in bookmarks:
            if not (bm.autojoin or config.getbool('open_all_bookmarks')):
                continue
            tab = self.tabs.by_name_and_class(bm.jid, MucTab)
            nick = bm.nick if bm.nick else self.own_nick
            if not tab:
                tab = self.open_new_room(
                    bm.jid, nick, focus=False, password=bm.password)
            self.initial_joins.append(bm.jid)
            # do not join rooms that do not have autojoin
            # but display them anyway
            if bm.autojoin and tab:
                tab.join()

    async def check_bookmark_storage(self, features: list[str]) -> None:
        """
        Check if remote bookmarkr storage is available.
        """
        private = 'jabber:iq:private' in features
        pep_ = 'http://jabber.org/protocol/pubsub#publish' in features
        self.bookmarks.available_storage['private'] = private
        self.bookmarks.available_storage['pep'] = pep_

        if not self.xmpp.anon and config.getbool('use_remote_bookmarks'):
            try:
                await self.bookmarks.get_remote(self.xmpp, self.information)
            except IqError as error:
                type_ = error.iq['error']['type']
                condition = error.iq['error']['condition']
                if not (type_ == 'cancel' and condition == 'item-not-found'):
                    self.information(
                        'Unable to fetch the remote'
                        ' bookmarks; %s: %s' % (type_, condition), 'Error')
                return
            remote_bookmarks = self.bookmarks.remote()
            self.join_initial_rooms(remote_bookmarks)

    def room_error(self, error: Presence | Message | Iq, room_name: str) -> None:
        """
        Display the error in the tab
        """
        tab = self.tabs.by_name_and_class(room_name, MucTab)
        if not tab:
            return
        error_message = get_error_message(error)
        tab.add_message(
            UIMessage(
                error_message,
                level='Error',
            ),
        )
        code = error['error']['code']
        if code == '401':
            msg = (
                'To provide a password in order to join the room, '
                'type "/join / password" (replace "password" by the '
                'real password)'
            )
            tab.add_message(PersistentInfoMessage(msg))
        if code == '409':
            if config.getstr('alternative_nickname') != '':
                if not tab.joined:
                    tab.own_nick += config.getstr('alternative_nickname')
                    tab.join()
            else:
                if not tab.joined:
                    tab.add_message(
                        PersistentInfoMessage(
                            'You can join the room with another nick, '
                            'by typing "/join /other_nick"'
                        )
                    )
        self.refresh_window()

    ################# Task management

    def register_task(self, coro: Coroutine[Any, Any, U], timeout: int | None = None) -> asyncio.Task[U]:
        if timeout is not None:
            task = asyncio.create_task(_timeout_task(coro, timeout))
        else:
            task = asyncio.create_task(coro)
        self._tasks.append(task)
        new_tasks = []
        for task in self._tasks[:]:
            if not task.done():
                new_tasks.append(task)
        self._tasks = new_tasks
        return task


async def _timeout_task(coro: Coroutine[Any, Any, U], timeout: int) -> U:
    async with asyncio.timeout(timeout):
        return await coro


class ScreenNotAvailable(ValueError):
    def __init__(self) -> None:
        super().__init__('No output available')


class KeyDictNotSetup(ValueError):
    def __init__(self) -> None:
        super().__init__("KeyDict not initialized")


class KeyDict(dict[str, Callable[..., Any]]):
    """
    A dict, with a wrapper for get() that will return a custom value
    if the key starts with _exc_
    """
    try_execute: Optional[Callable[[str], None]]

    def get_cb(self, key: str, default: Callable[..., Any] | None = None) -> Callable[..., Any] | None:
        if isinstance(key, str) and key.startswith('_exc_') and len(key) > 5:
            if self.try_execute is not None:
                try_execute = self.try_execute

                def partial_try(k: str = key[5:]) -> None:
                    try_execute(k)
                return partial_try
            raise KeyDictNotSetup()
        return dict.get(self, key, default)


def replace_key_with_bound(key: str) -> str:
    """
    Replace an inputted key with the one defined as its replacement
    in the config
    """
    return config.get(key, default=key, section='bindings') or key


def replace_line_breaks(key: str) -> str:
    "replace ^J with \n"
    if key == '^J':
        return '\n'
    return key


def separate_chars_from_bindings(char_list: List[str]) -> List[List[str]]:
    """
    returns a list of lists. For example if you give
    ['a', 'b', 'KEY_BACKSPACE', 'n', 'u'], this function returns
    [['a', 'b'], ['KEY_BACKSPACE'], ['n', 'u']]

    This way, in case of lag (for example), we handle the typed text
    by “batch” as much as possible (instead of one char at a time,
    which implies a refresh after each char, which is very slow),
    but we still handle the special chars (backspaces, arrows,
    ctrl+x ou alt+x, etc) one by one, which avoids the issue of
    printing them OR ignoring them in that case.  This should
    resolve the “my ^W are ignored when I lag ;(”.
    """
    res = []
    current = []
    for char in char_list:
        assert char
        # Transform that stupid char into what we actually meant
        if char == '\x1f':
            char = '^/'
        if len(char) == 1:
            current.append(char)
        else:
            # special case for the ^I key, it’s considered as \t
            # only when pasting some text, otherwise that’s the ^I
            # (or M-i) key, which stands for completion by default.
            if char == '^I' and len(char_list) != 1:
                current.append('\t')
                continue
            if current:
                res.append(current)
                current = []
            res.append([char])
    if current:
        res.append(current)
    return res



"""
Test the completions methods on an altered input object.
"""

import string
import pytest
import random
from poezio import config
from poezio.windows import Input


class ConfigShim(object):
    def get(self, *args, **kwargs):
        return ''
    def getbool(self, *args, **kwargs):
        return False


config.config = ConfigShim()


class SubInput(Input):
    def resize(self, *args, **kwargs):
        pass
    def rewrite_text(self, *args, **kwargs):
        pass
    def refresh(self, *args, **kwargs):
        pass


@pytest.fixture(scope="function")
def input_obj():
    from poezio.windows import base_wins
    base_wins.TAB_WIN = True
    obj = SubInput()
    obj.reset_completion()
    return obj

@pytest.fixture(scope="module")
def random_unquoted_words():
    letters = string.ascii_lowercase + ((len(string.ascii_lowercase)//4)*' ')
    acc = [random.choice(letters) for _ in range(200)]
    words = ''.join(acc).split()
    return words

@pytest.fixture(scope="module")
def quoted_words():
    words = []
    letters = string.ascii_lowercase + ((len(string.ascii_lowercase)//4)*' ')
    words_by_letter = {}
    for start_letter in string.ascii_lowercase:
        words_by_letter[start_letter] = []
        for _ in range(5):
            size = random.randint(0, 15)
            word = start_letter + ''.join(random.choice(letters) for i in range(size))
            words.append(word)
            words_by_letter[start_letter].append(word)
    return (words, words_by_letter)


def test_new_completion_1_unquoted(input_obj):

    input_obj.text = '/example '
    input_obj.pos = len(input_obj.text) - 1

    input_obj.new_completion(['toto', 'titi'], 1, quotify=False)
    assert input_obj.text == '/example toto'

    input_obj.new_completion(['toto', 'titi'], 1, quotify=False)
    assert input_obj.text == '/example titi'

    input_obj.new_completion(['toto', 'titi'], 1, quotify=False)
    assert input_obj.text == '/example toto'


def test_new_completion_1_quoted_spaces(input_obj):
    input_obj.text = '/example '
    input_obj.pos = len(input_obj.text) - 1

    input_obj.new_completion(['toto toto', 'titi titi'], 1, quotify=True)
    assert input_obj.text == '/example "toto toto"'

    input_obj.new_completion(['toto toto', 'titi titi'], 1, quotify=True)
    assert input_obj.text == '/example "titi titi"'

    input_obj.new_completion(['toto toto', 'titi titi'], 1, quotify=True)
    assert input_obj.text == '/example "toto toto"'

    input_obj.text = '/example '
    input_obj.pos = len(input_obj.text) - 1
    input_obj.reset_completion()

    input_obj.new_completion(['toto toto', 'tata', 'titi titi'], 1, quotify=True)
    assert input_obj.text == '/example "toto toto"'

    input_obj.new_completion(['toto toto', 'tata', 'titi titi'], 1, quotify=True)
    assert input_obj.text == '/example tata'

    input_obj.new_completion(['toto toto', 'tata', 'titi titi'], 1, quotify=True)
    assert input_obj.text == '/example "titi titi"'

def test_new_completion_unquoted_random_override(input_obj, random_unquoted_words):
    """
    Complete completely random words and ensure that the input is
    changed adequately.
    """
    words = random_unquoted_words

    # try the completion on the middle element without affecting the others
    input_obj.text = '/example %s %s %s' % (words[0], words[1], words[2])
    base = len(input_obj.text) - len(words[2]) - 1
    input_obj.pos = base
    def f(n):
        return '/example %s %s' % (words[0], words[n])

    for i in range(len(words)):
        pos = input_obj.get_argument_position(False)
        input_obj.new_completion(words[:], pos, quotify=False, override=True)
        assert f(i) + " " + words[2] == input_obj.text
        assert len(f(i)) == input_obj.pos

    assert input_obj.text == '/example %s %s %s' % (words[0], words[-1], words[2])

    pos = input_obj.get_argument_position(False)
    input_obj.new_completion(words[:], pos, quotify=False, override=True)
    assert input_obj.text == '/example %s %s %s' % (words[0], words[0], words[2])

    input_obj.reset_completion()

    # try the completion on the final element without affecting the others
    input_obj.text = '/example %s %s %s' % (words[0], words[1], words[2])
    base = len(input_obj.text)
    input_obj.pos = base
    def f2(n):
        return '/example %s %s %s' % (words[0], words[1], words[n])

    print(words)
    for i in range(len(words)):
        pos = input_obj.get_argument_position(False)
        input_obj.new_completion(words[:], pos, quotify=False, override=True)
        assert f2(i) == input_obj.text
        assert len(f2(i)) == input_obj.pos

    assert input_obj.text == '/example %s %s %s' % (words[0], words[1], words[-1])


def test_new_completion_quoted_random(input_obj, quoted_words):
    """
    Complete (possibly) quoted words starting with a specific letter.
    And make sure that the quotes only appear when necessary.
    """
    words = quoted_words[0]
    words_l = quoted_words[1]

    letters = ('', 'a', 'b', 'c')

    # generate the text which is supposed to be present in the input
    def f(p, i):
        rep = words_l[letters[p]][i] if ' ' not in words_l[letters[p]][i] else '"'+words_l[letters[p]][i]+'"'
        fst = letters[1] if p != 1 else rep
        snd = letters[2] if p != 2 else rep
        trd = letters[3] if p != 3 else rep
        return '/example %s %s %s' % (fst, snd, trd)

    for pos in range(1, 4):
        input_obj.text = '/example a b c'
        input_obj.reset_completion()
        input_obj.pos = len('/example') + pos * 2

        extra = (3 - pos) * 2
        for i in range(5):
            input_obj.new_completion(words[:], pos, quotify=True)
            assert f(pos, i) == input_obj.text
            assert len(f(pos, i)) - extra == input_obj.pos




"""
The command added by this plugin sends a message to someone when he next joins.

Usage
-----
This plugin defines two new commands for chatroom tabs:
:term:`/tell`, :term:`/untell`, and :term:`/list_tell`.

.. glossary::
    :sorted:

    /tell
         **Usage:** ``/tell <nick> <message>``

         Send *message* to *nick* at his next join.

    /untell
        **Usage:** ``/untell <nick>``

        Cancel all scheduled messages to *nick*.

    /list_tell
        **Usage:** ``/list_tell``

        List all queued messages for the current chatroom.

"""
import asyncio
from poezio.plugin import BasePlugin
from poezio.core.structs import Completion
from poezio.decorators import command_args_parser
from poezio.windows.inputs import Input
from poezio import tabs
from slixmpp import Presence


class Plugin(BasePlugin):
    def init(self) -> None:
        self.api.add_tab_command(
            tabs.MucTab,
            'tell',
            self.command_tell,
            usage='<nick> <message>',
            help='Will tell <nick> of <message> when he next joins.',
            short='Send a message when someone joins')
        self.api.add_tab_command(
            tabs.MucTab,
            'untell',
            self.command_untell,
            usage='<nick>',
            help='Remove the planned messages from /tell.',
            short='Cancel a /tell message',
            completion=self.completion_untell)
        self.api.add_tab_command(
            tabs.MucTab,
            'list_tell',
            self.command_list_tell,
            usage='',
            help='List currently queued messages')
        self.api.add_event_handler('muc_join', self.on_join)
        self.api.add_event_handler('muc_nickchange', self.on_join)
        # {tab -> {nick -> [messages]}
        self.tabs: dict[tabs.Tab, dict[str, list[str]]] = {}

    def on_join(self, presence: Presence, tab: tabs.ChatTab) -> None:
        if tab not in self.tabs:
            return
        nick = presence['from'].resource
        if nick not in self.tabs[tab]:
            return
        for i in self.tabs[tab][nick]:
            asyncio.ensure_future(tab.command_say("%s: %s" % (nick, i)))
        del self.tabs[tab][nick]

    @command_args_parser.ignored
    def command_list_tell(self) -> None:
        tab = self.api.current_tab()
        if not isinstance(tab, tabs.ChatTab):
            return
        if not self.tabs.get(tab):
            self.api.information('No message queued.', 'Info')
            return
        build = ['Messages queued for %s:' % tab.jid.bare]
        for nick, messages in self.tabs[tab].items():
            build.append(' for %s:' % nick)
            for message in messages:
                build.append(' - %s' % message)
        self.api.information('\n'.join(build), 'Info')

    @command_args_parser.quoted(2)
    def command_tell(self, args: list[str]) -> None:
        """/tell <nick> <message>"""
        if not args:
            self.core.command.help('tell')
            return
        nick, msg = args
        tab = self.api.current_tab()
        if tab not in self.tabs:
            self.tabs[tab] = {}
        if nick not in self.tabs[tab]:
            self.tabs[tab][nick] = []
        self.tabs[tab][nick].append(msg)
        self.api.information('Message for %s queued' % nick, 'Info')

    def command_untell(self, args: str) -> None:
        """/untell <nick>"""
        tab = self.api.current_tab()
        if tab not in self.tabs:
            return
        nick = args
        if nick not in self.tabs[tab]:
            return
        del self.tabs[tab][nick]
        self.api.information('Messages for %s unqueued' % nick, 'Info')

    def completion_untell(self, the_input: Input) -> Completion:
        tab = self.api.current_tab()
        if tab not in self.tabs:
            return Completion(the_input.auto_completion, [], '')
        return Completion(
            the_input.auto_completion, list(self.tabs[tab]), '', quotify=False)



# Copyright 2010-2011 Florent Le Coz <louiz@louiz.org>
#
# This file is part of Poezio.
#
# Poezio is free software: you can redistribute it and/or modify
# it under the terms of the GPL-3.0+ license. See the COPYING file.
"""
The logger module that handles logging of the poezio
conversations and roster changes
"""

import mmap
import re
from typing import IO, Generator, TypedDict
from datetime import datetime
from pathlib import Path

from poezio import common
from poezio.config import config
from poezio.xhtml import clean_text
from poezio.ui.types import Message, BaseMessage, LoggableTrait
from slixmpp import JID, InvalidJID

import logging

log = logging.getLogger(__name__)

MESSAGE_LOG_RE = re.compile(r'^MR (\d{4})(\d{2})(\d{2})T'
                            r'(\d{2}):(\d{2}):(\d{2})Z '
                            r'(\d+) <([^ ]+)>  (.*)$')
INFO_LOG_RE = re.compile(r'^MI (\d{4})(\d{2})(\d{2})T'
                         r'(\d{2}):(\d{2}):(\d{2})Z '
                         r'(\d+) (.*)$')


class LogItem:
    __slots__ = ('time', 'nb_lines', 'text')
    time: datetime
    nb_lines: int
    text: str

    def __init__(self, year: str, month: str, day: str, hour: str, minute: str,
                 second: str, nb_lines: str,
                 message: str):
        self.time = datetime(
            int(year), int(month), int(day), int(hour), int(minute),
            int(second))
        self.nb_lines = int(nb_lines)
        self.text = message


class LogInfo(LogItem):
    pass


class LogMessage(LogItem):
    __slots__ = ('nick', )
    nick: str

    def __init__(self, year: str, month: str, day: str, hour: str, minute: str,
                 seconds: str, nb_lines: str, nick: str,
                 message: str):
        LogItem.__init__(self, year, month, day, hour, minute, seconds,
                         nb_lines, message)
        self.nick = nick


LogDict = TypedDict(
    'LogDict',
    {
        'type': str, 'txt': str, 'time': datetime,
        'history': bool, 'nickname': str
    },
    total=False,
)


def parse_log_line(msg: str, jid: str = '') -> LogItem | None:
    """Parse a log line.

    :param msg: The message ligne
    :param jid: jid (for error logging)
    :returns: The LogItem or None on error
    """
    match = MESSAGE_LOG_RE.match(msg)
    if match:
        return LogMessage(*match.groups())
    match = INFO_LOG_RE.match(msg)
    if match:
        return LogInfo(*match.groups())
    log.debug('Error while parsing %s’s logs: “%s”', jid, msg)
    return None


class Logger:
    """
    Appends things to files. Error/information/warning logs
    and also log the conversations to logfiles
    """
    _roster_logfile: IO[str] | None
    log_dir: Path
    _fds: dict[str, IO[str]]
    _busy_fds: dict[str, bool]
    _buffered_fds: dict[str, list[str]]

    def __init__(self) -> None:
        self.log_dir = Path()
        self._roster_logfile = None
        # a dict of 'groupchatname': file-object (opened)
        self._fds = {}
        self._busy_fds = {}
        self._buffered_fds = {}

    def __del__(self) -> None:
        """Close all fds on exit"""
        for jid, opened_file in self._fds.items():
            if opened_file:
                try:
                    opened_file.close()
                except Exception:  # Can't close? too bad
                    log.debug('Unable to close a file for %s', jid)
        if self._roster_logfile:
            try:
                self._roster_logfile.close()
            except Exception:
                log.debug('Unable to close the roster log')

    def get_file_path(self, jid: str | JID) -> Path:
        """Return the log path for a specific jid"""
        jidstr = str(jid).replace('/', '\\')
        return self.log_dir / jidstr

    def fd_busy(self, jid: str | JID) -> None:
        """Signal to the logger that this logfile is busy elsewhere.
        And that the messages should be queued to be logged later.

        :param jid: file name
        """
        jidstr = str(jid).replace('/', '\\')
        self._busy_fds[jidstr] = True
        if jidstr not in self._buffered_fds:
            self._buffered_fds[jidstr] = []

    def fd_available(self, jid: str | JID) -> None:
        """Signal to the logger that this logfile is no longer busy.
        And write messages to the end.

        :param jid: file name
        """
        jidstr = str(jid).replace('/', '\\')
        if jidstr in self._busy_fds:
            del self._busy_fds[jidstr]
        if jidstr in self._buffered_fds:
            msgs = ''.join(self._buffered_fds.pop(jidstr))
            if jidstr in self._fds:
                self._fds[jidstr].close()
                del self._fds[jidstr]
            self.log_raw(jid, msgs)

    def close(self, jid: str) -> None:
        """Close the log file for a JID."""
        jidstr = str(jid).replace('/', '\\')
        if jidstr in self._fds:
            self._fds[jidstr].close()
            log.debug('Log file for %s closed.', jid)
            del self._fds[jidstr]

    def reload_all(self) -> None:
        """Close and reload all the file handles (on SIGUSR1)"""
        not_closed = set()
        for key, opened_file in self._fds.items():
            if opened_file:
                try:
                    opened_file.close()
                except Exception:
                    not_closed.add(key)
        if self._roster_logfile:
            try:
                self._roster_logfile.close()
            except Exception:
                not_closed.add('roster')
        log.debug('All log file handles closed')
        if not_closed:
            log.error('Unable to close log files for: %s', not_closed)
        for room in list(self._fds):
            self._fds.pop(room, None)
            if self._check_and_create_log_dir(room) is not None:
                log.debug('Log handle for %s re-created', room)

    def _check_and_create_log_dir(self, jid: str | JID,
                                  open_fd: bool = True) -> IO[str] | None:
        """
        Check that the directory where we want to log the messages
        exists. if not, create it

        :param jid: JID of the file to open after creating the dir
        :param open_fd: if the file should be opened after creating the dir
        :returns: the opened fd or None
        """
        # POSIX filesystems don't support / in filename, so we replace it with a backslash
        strjid = str(jid).replace('/', '\\')
        try:
            jid = JID(jid)
            if not config.get_by_tabname('use_log', jid):
                return None
            self.log_dir.mkdir(parents=True, exist_ok=True)
            if open_fd:
                filename = self.get_file_path(strjid)
                fd = filename.open('a', encoding='utf-8')
                self._fds[strjid] = fd
                return fd
        except InvalidJID:
            log.debug('Unable to convert %s to JID', jid)
            return None
        except IOError:
            log.exception('Unable to open the log file (%s)', filename)
        except Exception:
            log.exception('Unable to create the log dir')
            return None
        return None

    def log_message(self,
                    jid: str,
                    msg: BaseMessage | Message) -> bool:
        """
        Log the message in the appropriate file

        :param jid: JID of the entity for which to log the message
        :param msg: Message to log
        :returns: True if no error was encountered
        """
        if not config.get_by_tabname('use_log', JID(jid)):
            return True
        if not isinstance(msg, LoggableTrait):
            return True
        date = msg.time
        txt = msg.txt
        nick = ''
        typ = 'MI'
        if isinstance(msg, Message):
            nick = msg.nickname or ''
            if msg.me:
                txt = f'/me {txt}'
            typ = 'MR'
        logged_msg = build_log_message(nick, txt, date=date, prefix=typ)
        if not logged_msg:
            return True
        return self.log_raw(jid, logged_msg)

    def log_raw(self, jid: str | JID, logged_msg: str, force: bool = False) -> bool:
        """Log a raw string.

        :param jid: filename
        :param logged_msg: string to log
        :param force: Bypass the buffered fd check
        :returns: True if no error was encountered
        """
        jidstr = str(jid).replace('/', '\\')
        if jidstr in self._fds.keys():
            fd = self._fds[jidstr]
        else:
            option_fd = self._check_and_create_log_dir(jid)
            if option_fd is None:
                return True
            fd = option_fd
        filename = self.get_file_path(jid)
        try:
            if not force and self._busy_fds.get(jidstr):
                self._buffered_fds[jidstr].append(logged_msg)
                return True
            fd.write(logged_msg)
        except (OSError, ValueError):
            log.exception(
                'Unable to write in the log file (%s)',
                filename,
            )
            return False
        else:
            try:
                fd.flush()  # TODO do something better here?
            except (OSError, ValueError):
                log.exception(
                    'Unable to flush the log file (%s)',
                    filename,
                )
                return False
        return True

    def log_roster_change(self, jid: str, message: str) -> bool:
        """
        Log a roster change

        :param jid: jid to log the change for
        :param message: message to log
        :returns: True if no error happened
        """
        if not config.get_by_tabname('use_log', JID(jid)):
            return True
        self._check_and_create_log_dir('', open_fd=False)
        filename = self.log_dir / 'roster.log'
        if not self._roster_logfile:
            try:
                self._roster_logfile = filename.open('a', encoding='utf-8')
            except IOError:
                log.exception(
                    'Unable to create the log file (%s)',
                    filename,
                )
                return False
        try:
            str_time = common.get_utc_time().strftime('%Y%m%dT%H:%M:%SZ')
            message = clean_text(message)
            lines = message.split('\n')
            first_line = lines.pop(0)
            nb_lines = str(len(lines)).zfill(3)
            self._roster_logfile.write(
                'MI %s %s %s %s\n' % (str_time, nb_lines, jid, first_line))
            for line in lines:
                self._roster_logfile.write(' %s\n' % line)
            self._roster_logfile.flush()
        except Exception:
            log.exception(
                'Unable to write in the log file (%s)',
                filename,
            )
            return False
        return True


def build_log_message(nick: str | None,
                      msg: str,
                      date: datetime | None = None,
                      prefix: str = 'MI') -> str:
    """
    Create a log message from a nick, a message, optionally a date and type

    :param nick: nickname to log
    :param msg: text of the message
    :param date: date of the message
    :param prefix: MI (info) or MR (message)
    :returns: The log line(s)
    """
    msg = clean_text(msg)
    time = common.get_utc_time() if date is None else common.get_utc_time(date)
    str_time = time.strftime('%Y%m%dT%H:%M:%SZ')
    lines = msg.split('\n')
    first_line = lines.pop(0)
    nb_lines = str(len(lines)).zfill(3)
    if nick:
        nick = '<' + nick + '>'
        logged_msg = '%s %s %s %s  %s\n' % (prefix, str_time, nb_lines, nick,
                                            first_line)
    else:
        logged_msg = '%s %s %s %s\n' % (prefix, str_time, nb_lines, first_line)
    return logged_msg + ''.join(' %s\n' % line for line in lines)


def last_message_in_archive(filepath: Path) -> LogDict | None:
    """Get the last message from the local archive.

    :param filepath: the log file path
    """
    last_msg = None
    for msg in iterate_messages_reverse(filepath):
        if msg['type'] == 'message':
            last_msg = msg
            break
    return last_msg


def iterate_messages_reverse(filepath: Path) -> Generator[LogDict, None, None]:
    """Get the latest messages from the log file, one at a time.

    :param fd: the file descriptor
    """
    try:
        with open(filepath, 'rb') as fd:
            with mmap.mmap(fd.fileno(), 0, prot=mmap.PROT_READ) as m:
                # start of messages begin with MI or MR, after a \n
                pos = m.rfind(b"\nM") + 1
                lines = []
                if pos != -1:
                    lines = parse_log_lines(
                        m[pos:-1].decode(errors='replace').splitlines()
                    )
                elif m[0:1] == b'M':
                    # Handle the case of a single message present in the log
                    # file, hence no newline.
                    lines = parse_log_lines(
                        m[:].decode(errors='replace').splitlines()
                    )
                if lines:
                    yield lines[0]
                while pos > 0:
                    old_pos = pos
                    pos = m.rfind(b"\nM", 0, pos) + 1
                    lines = parse_log_lines(
                        m[pos:old_pos].decode(errors='replace').splitlines()
                    )
                    if lines:
                        yield lines[0]
    except (OSError, ValueError):
        pass


def parse_log_lines(lines: list[str], jid: str = '') -> list[LogDict]:
    """
    Parse raw log lines into poezio log objects

    :param lines: Message lines
    :param jid: jid (for error logging)
    :return: a list of dicts containing message info
    """
    messages = []

    # now convert that data into actual Message objects
    idx = 0
    while idx < len(lines):
        if lines[idx].startswith(' '):  # should not happen ; skip
            idx += 1
            log.debug('fail?')
            continue
        log_item: LogItem | None = parse_log_line(lines[idx], jid)
        idx += 1
        if not isinstance(log_item, LogItem):
            log.debug('wrong log format? %s', log_item)
            continue
        message_lines = []
        message = LogDict({
            'history': True,
            'time': common.get_local_time(log_item.time),
            'type': 'message',
        })
        size = log_item.nb_lines
        if isinstance(log_item, LogInfo):
            message_lines.append(log_item.text)
            message['type'] = 'info'
        elif isinstance(log_item, LogMessage):
            message['nickname'] = log_item.nick
            message_lines.append(log_item.text)
        while size != 0 and idx < len(lines):
            message_lines.append(lines[idx][1:])
            size -= 1
            idx += 1
        message['txt'] = '\n'.join(message_lines)
        messages.append(message)
    return messages


logger = Logger()



"""
This plugin uses figlet to transform every message into a big ascii-art
message.


Usage
-----

Say something in a Chat tab.

.. note:: Can create fun things when used with :ref:`The rainbow plugin <rainbow-plugin>`.

"""

import subprocess
from poezio.plugin import BasePlugin
from slixmpp import Message
from typing import Any


def is_figlet() -> bool:
    """Ensure figlet exists"""
    process = subprocess.Popen(
        ['which', 'figlet'],
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    return process.wait() == 0


class Plugin(BasePlugin):
    def init(self) -> None:
        if not is_figlet():
            self.api.information(
                'Couldn\'t find the figlet program. '
                'Please install it and reload the plugin.',
                'Error',
            )
            return None

        self.api.add_event_handler('muc_say', self.figletize)
        self.api.add_event_handler('conversation_say', self.figletize)
        self.api.add_event_handler('private_say', self.figletize)
        return None

    def figletize(self, msg: Message, tab: Any) -> None:
        process = subprocess.Popen(
            ['figlet', '--', msg['body']], stdout=subprocess.PIPE)
        result = process.communicate()[0].decode('utf-8')
        msg['body'] = result
